"""Hydra configuration classes for CryoViT experiments."""

import logging
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Any

from hydra.core.config_store import ConfigStore
from omegaconf import MISSING, OmegaConf

from cryovit.types import Sample

samples: list[str] = [sample.name for sample in Sample]
tomogram_exts: list[str] = [".hdf", ".mrc"]

DINO_PATCH_SIZE = 14
DEFAULT_WINDOW_SIZE = (
    630  # Calculated assuming DINOv2 patch size of 14 and input size of 720
)


@dataclass
class BaseModel:
    """Base configuration for models used in CryoViT experiments.

    Attributes:
        name (str): Name of the model for identification purposes.
        input_key (str): Key to get the input data from a tomogram.
        model_dir (Optional[Path]): Optional directory to download model weights to (for SAMv2 models).
        lr (float): Learning rate for the model training.
        weight_decay (float): Weight decay (L2 penalty) rate. Default is 1e-3.
        losses (dict[str, Any]): Configurations for loss functions used in training.
        metrics (dict[str, Any]): Configurations for metrics used during model evaluation.
        custom_kwargs (Optional[dict[str, Any]]): Optional dictionary of custom keyword arguments to pass to the model.
    """

    _target_: str = MISSING
    name: str = MISSING

    input_key: str = MISSING
    model_dir: Path | None = None
    lr: float = MISSING
    weight_decay: float = 1e-3
    losses: dict = MISSING
    metrics: dict = MISSING

    custom_kwargs: dict | None = None


@dataclass
class BaseTrainer:
    """Base configuration for the trainer used in CryoViT experiments.

    Attributes:
        accelerator (str): Type of hardware acceleration. Default is 'gpu'.
        devices (str): Number of devices to use for training. Default is '1'.
        precision (str): Precision configuration for training (e.g., '16-mixed').
        default_root_dir (Optional[Path]): Default root directory for saving checkpoints and logs.
        max_epochs (Optional[int]): The maximum number of epochs to train for.
        enable_checkpointing (bool): Flag to enable or disable model checkpointing. Default is False.
        enable_model_summary (bool): Enable model summarization. Default is True.
        log_every_n_steps (Optional[int]): Frequency of logging in terms of training steps.
    """

    _target_: str = "pytorch_lightning.Trainer"

    accelerator: str = "gpu"
    devices: str = "1"
    precision: str = "16-mixed"
    default_root_dir: Path | None = None
    max_epochs: int | None = None
    enable_checkpointing: bool = False
    enable_model_summary: bool = True
    log_every_n_steps: int | None = None


@dataclass
class BaseDataModule:
    """Base configuration for datasets in CryoViT experiments.

    Attributes:
        sample (Union[Sample, tuple[Sample]]): Specific sample or samples used for training.
        split_id (Optional[int]): Optional split_id to use for validation.
        split_key (Optional[str]): Key in the sample .csv file to use for splitting the data. Default is "split_id".
        test_sample (Optional[Any]): Specific sample or samples used for testing.
        dataset (dict[str, Any]): Configuration for the dataset.
        dataloader (dict[str, Any]): Configuration for the dataloader.
    """

    _target_: str = ""
    _partial_: bool = True

    # OmegaConf doesn't support Union[Sample, tuple[Sample]] yet, so moved type-checking to config validation instead
    sample: Any = MISSING
    split_id: int | None = None
    split_key: str | None = "split_id"
    test_sample: Any | None = None

    dataset: dict = MISSING
    dataloader: dict = MISSING


@dataclass
class ExperimentPaths:
    """Configuration for managing experiment paths in CryoViT experiments.

    Attributes:
        model_dir (Path): Path to the folder containing downloaded models.
        data_dir (Path): Path to the parent directory containing tomogram data and .csv files.
        exp_dir (Path): Path to the parent directory for saving results from an experiment.
        results_dir (Path): Path to the parent directory for saving overall results.
        tomo_name (str): Name of the folder in data_dir with tomograms.
        feature_name (str): Name of the folder in data_dir with DINOv2 features.
        dino_name (str): Name of the folder in model_dir to save DINOv2 model.
        csv_name (str): Name of the folder in data_dir with .csv files.
        split_name (str): Name of the .csv file with training splits.
    """

    model_dir: Path = MISSING
    data_dir: Path = MISSING
    exp_dir: Path = MISSING
    results_dir: Path = MISSING

    tomo_name: str = "tomograms"
    feature_name: str = "dino_features"
    dino_name: str = "DINOv2"
    sam_name: str = "SAM2"
    csv_name: str = "csv"
    split_name: str = "splits.csv"


@dataclass
class DinoFeaturesConfig:
    """Base configuration for computing DINOv2 features in CryoViT experiments.

    Attributes:
        batch_size (int): Number of tomogram slices to process as one batch. Default is 128.
        dino_dir (Path): Path to the DINOv2 foundation model.
        paths (ExperimentPaths): Configuration for experiment paths.
        datamodule (dict[str, Any]): Configuration for the datamodule to use for loading tomograms.
        sample (Optional[Sample]): Sample to calculate features for. None means to calculate features for all samples.
        export_features (bool): Whether to additionally compute PCA colormaps for the calculated features.
    """

    batch_size: int = 128
    dino_dir: Path = MISSING
    paths: ExperimentPaths = MISSING
    datamodule: dict = MISSING
    sample: Sample | None = MISSING
    export_features: bool = False


@dataclass
class BaseExperimentConfig:
    """Base configuration for running CryoViT experiments.

    Attributes:
        name (str): Name of the experiment, should be unique for each configuration.
        label_key (str): Key used to specify the training labels.
        additional_keys (tuple[str]): Keys to pass through additional data from the dataset.
        random_seed (int): Random seed set for reproducibility. Default is 42.
        paths (ExperimentPaths): Configuration for experiment paths.
        model (BaseModel): Configuration for the model to use.
        trainer (BaseTrainer): Configuration for the trainer to use.
        callbacks (Optional[list]): List of callback functions for the trainer.
        logger (Optional[list]): List of logging functions for the trainer.
        datamodule (BaseDataModule): Configuration for the datamodule to use.
        ckpt_path (Optional[Path]): Optional path to a checkpoint file to resume training from.
        resume_ckpt (bool): Whether to resume training from the checkpoint. Default is False.
    """

    name: str = MISSING
    label_key: str = MISSING
    additional_keys: tuple[str] = ()  # type: ignore
    random_seed: int = 42
    paths: ExperimentPaths = MISSING
    model: BaseModel = MISSING
    trainer: BaseTrainer = MISSING
    callbacks: dict[str, Any] = MISSING
    logger: dict[str, Any] = MISSING
    datamodule: BaseDataModule = MISSING
    ckpt_path: Path | None = None
    resume_ckpt: bool = False


cs = ConfigStore.instance()

cs.store(group="model", name="base_model", node=BaseModel)
cs.store(group="trainer", name="base_trainer", node=BaseTrainer)
cs.store(group="datamodule", name="base_datamodule", node=BaseDataModule)
cs.store(group="paths", name="base_env", node=ExperimentPaths)

cs.store(name="dino_features_config", node=DinoFeaturesConfig)
cs.store(name="base_experiment_config", node=BaseExperimentConfig)

#### Utility Functions for Configs ####\


def validate_dino_config(cfg: DinoFeaturesConfig) -> None:
    """Validates the configuration for DINOv2 feature extraction.

    Checks if all necessary parameters are present in the configuration. If any required parameters are
    missing, it logs an error message and exits the script.

    Args:
        cfg (DinoFeaturesConfig): The configuration object containing settings for feature extraction.

    Raises:
        SystemExit: If any configuration parameters are missing.
    """
    missing_keys = OmegaConf.missing_keys(cfg)
    error_msg = [
        "The following parameters were missing from dino_features.yaml"
    ]

    for i, key in enumerate(missing_keys, 1):
        param_dict = DinoFeaturesConfig.__annotations__
        error_str = f"{i}. {key}: {param_dict.get(key, Any).__name__}"
        error_msg.append(error_str)

    if missing_keys:
        logging.error("\n".join(error_msg))
        sys.exit(1)

    OmegaConf.set_struct(cfg, False)  # type: ignore


def validate_experiment_config(cfg: BaseExperimentConfig) -> None:
    """Validates an experiment configuration.

    Checks if all necessary parameters are present in the configuration. If any required parameters are
    missing, it logs an error message and exits the script.

    Additionally, checks that all Samples specified are valid, and logs an error and exits if any samples are not valid.

    Args:
        cfg (BaseExperimentConfig): The configuration object to validate.

    Raises:
        SystemExit: If any configuration parameters are missing, or any samples are not valid, terminating the script.
    """
    missing_keys = OmegaConf.missing_keys(cfg)
    error_msg = ["The following parameters were missing from config:"]

    for i, key in enumerate(missing_keys, 1):
        error_msg.append(f"{i}. {key}")

    if missing_keys:
        logging.error("\n".join(error_msg))
        sys.exit(1)

    # Check datamodule samples are valid
    error_msg = ["The following datamodule parameters are not valid samples:"]
    invalid_samples = []
    if isinstance(cfg.datamodule.sample, str):
        cfg.datamodule.sample = [cfg.datamodule.sample]
    if isinstance(cfg.datamodule.test_sample, str):
        cfg.datamodule.test_sample = [cfg.datamodule.test_sample]

    for sample in cfg.datamodule.sample:
        if sample not in samples:
            invalid_samples.append(sample)

    if cfg.datamodule.test_sample is not None and isinstance(
        cfg.datamodule.test_sample, list
    ):
        for sample in cfg.datamodule.test_sample:
            if sample not in samples:
                invalid_samples.append(sample)

    for i, sample in enumerate(invalid_samples, 1):
        error_msg.append(f"{i}. {sample}")

    if invalid_samples:
        logging.error("\n".join(error_msg))
        sys.exit(1)

    OmegaConf.set_struct(cfg, False)  # type: ignore



"""Dataset class for loading tomograms for CryoViT experiments."""

from pathlib import Path
from typing import Any

import h5py
import numpy as np
import pandas as pd
from torch.utils.data import Dataset

from cryovit.types import TomogramData


class TomoDataset(Dataset):
    """A dataset class for handling and preprocessing tomographic data for CryoVIT models."""

    def __init__(
        self,
        records: pd.DataFrame,
        input_key: str,
        label_key: str,
        split_key: str,
        data_root: Path,
        aux_keys: list[str] | None = None,
        train: bool = False,
    ) -> None:
        """Initializes a dataset object to load tomograms for model training, applying optional training crops.

        Args:
            records (pd.DataFrame): A DataFrame containing records of tomograms.
            input_key (str): The key in the HDF5 file to access input features.
            label_key (str): The key in the HDF5 file to access labels.
            split_key (str): The key in the DataFrame to access the split identifier.
            data_root (Path): The root directory where the tomograms are stored.
            aux_keys (Optional[List[str]]): Optional additional keys for auxiliary data to load from the HDF5 files.
            train (bool): Flag to determine if the dataset is for training (enables transformations).
        """

        if aux_keys is None:
            aux_keys = []
        self.records = records
        self.input_key = input_key
        self.label_key = label_key
        self.split_key = split_key
        self.aux_keys = aux_keys
        self.data_root = (
            data_root if isinstance(data_root, Path) else Path(data_root)
        )
        self.train = train

    def __len__(self) -> int:
        """Returns the total number of tomograms in the dataset."""

        return len(self.records)

    def __getitem__(self, idx: int) -> TomogramData:  # type: ignore
        """Retrieves a single item from the dataset.

        Args:
            idx (int): The index of the item.

        Returns:
            TomogramData: A dataclass containing the loaded data, labels, and metadata.

        Raises:
            IndexError: If index is out of the range of the dataset.
        """

        if idx >= len(self):
            raise IndexError

        record = self.records.iloc[idx]
        assert isinstance(record, pd.Series)
        data = self._load_tomogram(record)

        if self.train:
            self._random_crop(data)

        return TomogramData(
            sample=record["sample"],
            tomo_name=record["tomo_name"],
            split_id=data.get("split_id", None),
            data=data["input"],
            label=data["label"],
            aux_data={key: data[key] for key in self.aux_keys if key in data},
        )  # type: ignore

    def _load_tomogram(self, record: pd.Series) -> dict[str, Any]:
        """Loads a single tomogram based on the record information.

        Args:
            record (pd.Series): A series containing the sample and tomogram names.

        Returns:
            data (dict[str, Any]): A dictionary with input data, label, and any auxiliary data.
        """

        tomo_path = self.data_root / record["sample"] / record["tomo_name"]

        # sample, tomo_name, split_id, input, label

        data_dict = {
            "sample": record["sample"],
            "tomo_name": record["tomo_name"],
        }
        if self.split_key in record.index:
            data_dict["split_id"] = record[self.split_key]

        with h5py.File(tomo_path) as fh:
            assert (
                self.input_key in fh
            ), f"Input key '{self.input_key}' not found in {tomo_path}."
            assert (
                self.label_key in fh["labels"]
            ), f"Label key '{self.label_key}' not found in {tomo_path}/labels."  # type: ignore
            data: np.ndarray = fh[self.input_key][()]  # type: ignore
            if data.dtype == np.uint8:
                data = data.astype(np.float32) / 255.0
            if len(data.shape) == 3:
                data = data[np.newaxis, ...]  # add channel dimension
            data_dict["input"] = data
            data_dict["label"] = fh["labels"][self.label_key][()]  # type: ignore
            data_dict |= {key: fh[key][()] for key in self.aux_keys if key in fh}  # type: ignore

        return data_dict

    def _random_crop(self, data: dict[str, Any]) -> None:
        """Applies a random crop to the input data in the record dictionary.

        Args:
            record (dict[str, Any]): The record dictionary containing 'input' and 'label' data.
        """

        max_depth = 128
        side = 32 if self.input_key == "dino_features" else 512
        d, h, w = data["input"].shape[-3:]
        x, y, z = min(d, max_depth), side, side

        if (d, h, w) == (x, y, z):
            return  # nothing to be done

        delta_d = d - x + 1
        delta_h = h - y + 1
        delta_w = w - z + 1

        di = np.random.choice(delta_d) if delta_d > 0 else 0
        hi = np.random.choice(delta_h) if delta_h > 0 else 0
        wi = np.random.choice(delta_w) if delta_w > 0 else 0

        data["input"] = data["input"][
            ..., di : di + x, hi : hi + y, wi : wi + z
        ]

        if self.input_key == "dino_features":
            hi, wi, y, z = 16 * np.array([hi, wi, y, z])

        data["label"] = data["label"][di : di + x, hi : hi + y, wi : wi + z]



"""Script to evaluate segmentation models for CryoET data."""

import traceback
import warnings

import hydra

from cryovit._logging_config import setup_logging
from cryovit.config import BaseExperimentConfig, validate_experiment_config
from cryovit.run import eval_model

setup_logging("DEBUG")
import logging  # noqa: E402

warnings.simplefilter("ignore")


@hydra.main(
    config_path="../configs",
    config_name="eval_model.yaml",
    version_base="1.2",
)
def main(cfg: BaseExperimentConfig) -> None:
    """Main function to orchestrate the evaluation of segmentation models.

    First, it validates the configuration and, if valid, proceeds to run the evaluation process using
    the specified settings in the config file. It captures and logs any errors encountered during the
    evaluation process.

    Args:
        cfg (EvalModelConfig): Configuration object loaded from eval_model.yaml.
    """

    validate_experiment_config(cfg)

    try:
        eval_model.run_trainer(cfg)
    except BaseException as err:  # noqa: BLE001
        logging.error("%s: %s", type(err).__name__, err)
        logging.error(traceback.format_exc())


if __name__ == "__main__":
    main()



"""Defines the command-line interface (CLI) for CryoViT.

Copied from https://github.com/teamtomo/fidder/blob/main/src/fidder/_cli.py."""

import typer
from click import Context
from typer.core import TyperGroup


class OrderCommands(TyperGroup):
    """Return list of commands in the order appear."""

    def list_commands(self, ctx: Context):
        """Return list of commands in the order appear."""
        return list(self.commands)  # get commands using self.commands


cli = typer.Typer(
    cls=OrderCommands,
    add_completion=False,
    no_args_is_help=True,
    rich_markup_mode="rich",
    pretty_exceptions_show_locals=False,
)


@cli.callback()
def callback():
    """
    CryoViT's command line interface (CLI) for training, evaluation, and inference.

    You can choose between the different options listed below.
    To see the help for a specific command, run:

    cryovit <command> --help
    """



"""Utility functions to process data and models in a format recognizable by CryoVIT."""

import logging
import pickle
import random
import string
from dataclasses import dataclass
from pathlib import Path
from typing import Any

import h5py
import mrcfile
import numpy as np
import tifffile as tf
import torch
from hydra.utils import instantiate

from cryovit.config import BaseModel, tomogram_exts
from cryovit.models.sam2 import create_sam_model_from_weights
from cryovit.types import ModelType

#### General File Utilities ####

# Recognized file extensions for data files
RECOGNIZED_FILE_EXTS = [
    ".h5",
    ".hdf",
    ".hdf5",
    ".mrc",
    ".mrcs",
    ".tiff",
    ".tif",
]


def id_generator(size: int = 6, chars=string.ascii_lowercase + string.digits):
    """Generates a random string of fixed size."""

    return "".join(random.choice(chars) for _ in range(size))


#### Data Loading Utilities ####


@dataclass
class FileMetadata:
    """Metadata information for a file.

    Attributes:
        drange: The dynamic range of the data.
        dshape: The shape of the data.
        dtype: The data type of the data.
        nunique: The number of unique values in the data.
    """

    drange: tuple[float, float]
    dshape: tuple[int, ...]
    dtype: np.dtype
    nunique: int = 0


def _read_hdf_keys(
    hdf_file: h5py.File | h5py.Group, data_key: str | None = None
) -> tuple[dict[str, np.ndarray], dict[str, FileMetadata]]:
    """Recursively read all keys in an HDF5 file or group, returning as a tuple of dictionaries of key: <data>/<metadata>. If a data_key is specified, only reads that key."""

    if len(hdf_file.keys()) == 0:
        return {}, {}
    data_results = {}
    metadata_results = {}
    if data_key is not None:
        try:
            data: np.ndarray = hdf_file[data_key][()]  # type: ignore
            drange = (float(np.min(data)), float(np.max(data)))
            dshape = data.shape
            dtype = data.dtype
            nunique = len(np.unique(data))
            data_results[data_key] = data
            metadata_results[data_key] = FileMetadata(
                drange=drange, dshape=dshape, dtype=dtype, nunique=nunique
            )
            return data_results, metadata_results
        except KeyError:
            logging.warning(
                "Key %s not found in file %s. Attempting to read all keys instead.",
                data_key,
                hdf_file,
            )
    for key in hdf_file:
        if isinstance(hdf_file[key], h5py.Group):
            group_data_results, group_metadata_results = _read_hdf_keys(hdf_file[key])  # type: ignore
            data_results.update(
                {f"{key}/{k}": v for k, v in group_data_results.items()}
            )
            metadata_results.update(
                {f"{key}/{k}": v for k, v in group_metadata_results.items()}
            )
        elif isinstance(hdf_file[key], h5py.Dataset):
            data: np.ndarray = hdf_file[key][()]  # type: ignore
            drange = (float(np.min(data)), float(np.max(data)))
            dshape = data.shape
            dtype = data.dtype
            nunique = len(np.unique(data))
            data_results[key] = data
            metadata_results[key] = FileMetadata(
                drange=drange, dshape=dshape, dtype=dtype, nunique=nunique
            )
        else:
            raise ValueError(
                f"Unknown HDF5 object type found for key {key} in file {hdf_file.name}: {type(hdf_file[key])}."
            )
    return data_results, metadata_results


def read_hdf(
    hdf_file: str | Path, key: str | None = None
) -> tuple[str, np.ndarray, FileMetadata]:
    """Read data from an HDF5 file. If a key is not specified, assumes the data with the most unique values is the data.

    Args:
        hdf_file: The path to the HDF5 file.
        key: The key to read from the HDF5 file. If None, assumes the data with the most unique values is the data. If not a valid key, will attempt to read all keys and use the one with the most unique values.

    Returns:
        A tuple of the key used, the data, and the metadata.
    """

    with h5py.File(hdf_file, "r") as f:
        data_dict, metadata_dict = _read_hdf_keys(f, data_key=key)
    if key is None:
        # Assume the data with the most unique values is the data
        data_key = max(metadata_dict.items(), key=lambda x: x[1].nunique)[0]
        logging.info(
            "No key specified for file %s. Assuming data is the key with the most unique values, and using key '%s' with %d unique values. If this is incorrect, please specify the `data_key` manually as a `/`-separated string.",
            hdf_file,
            data_key,
            metadata_dict[data_key].nunique,
        )
    else:
        data_key = key
    data = data_dict[data_key]
    metadata = metadata_dict[data_key]
    return data_key, data, metadata


def read_mrc(mrc_file: str | Path) -> tuple[np.ndarray, FileMetadata]:
    """Read data from an MRC file.

    Args:
        mrc_file: The path to the MRC file.

    Returns:
        A tuple of the data and the metadata.
    """

    data: np.ndarray = mrcfile.read(mrc_file)
    drange = (float(np.min(data)), float(np.max(data)))
    dshape = data.shape
    dtype = data.dtype
    nunique = len(np.unique(data))
    return data, FileMetadata(
        drange=drange, dshape=dshape, dtype=dtype, nunique=nunique
    )


def read_tiff(tiff_file: str | Path) -> tuple[np.ndarray, FileMetadata]:
    """Read data from a TIFF file.

    Args:
        tiff_file: The path to the TIFF file.

    Returns:
        A tuple of the data and the metadata.
    """

    data: np.ndarray = tf.imread(tiff_file)
    drange = (float(np.min(data)), float(np.max(data)))
    dshape = data.shape
    dtype = data.dtype
    nunique = len(np.unique(data))
    return data, FileMetadata(
        drange=drange, dshape=dshape, dtype=dtype, nunique=nunique
    )


def load_data(
    file_path: str | Path, key: str | None = None
) -> tuple[np.ndarray, str]:
    """Load data or labels from a given file path. Supports .h5, .hdf5, .mrc, .mrcs formats.

    Args:
        file_path: The path to the data file.
        key: An optional key to specify which dataset to load from an HDF5 file.

    Raises:
        ValueError: If the file format is unsupported.
        FileNotFoundError: If the specified file does not exist.

    Returns:
        A tuple of the data and the key used (empty string if not applicable).
    """

    file_path = Path(file_path)
    found_key = ""
    if not file_path.exists():
        raise FileNotFoundError(f"File {file_path} does not exist.")
    if file_path.suffix in [".h5", ".hdf", ".hdf5"]:
        found_key, data, metadata = read_hdf(file_path, key=key)
    elif file_path.suffix in [".mrc", ".mrcs"]:
        data, metadata = read_mrc(file_path)
    elif file_path.suffix in [".tiff", ".tif"]:
        data, metadata = read_tiff(file_path)
    else:
        raise ValueError(
            f"Unsupported file format for file {file_path}. Supported formats are .h5, .hdf, .hdf5, .mrc, .mrcs, .tiff, .tif, and image folders."
        )

    # Assumed float data already normalized or are DINO features
    if metadata.dtype in [np.uint8, np.int8, np.uint16, np.int16]:
        # Normalize to [0, 1] and return as float32
        data = data.astype(np.float32) / 255.0

    if len(data.shape) == 3:
        data = data[np.newaxis, ...]  # add channel dimension
    return data, found_key


def _match_label_keys_to_data(
    data: np.ndarray, label_keys: list[str], metadata: FileMetadata
) -> dict[str, np.ndarray]:
    """Match label keys to data based on the number of unique values in the data and the provided label keys. Assumes label_keys are in ascending-value order."""

    labels = {}
    nunique = (
        metadata.nunique if metadata.drange[0] >= 0 else metadata.nunique - 1
    )  # ignore negative values for nunique count
    if nunique == len(label_keys):
        label_values = sorted(np.unique(data).tolist())
        for i, key in zip(label_values, label_keys, strict=True):
            label = np.where((data != i) & (data != -1), 0, data)
            labels[key] = np.where(label == i, 1, label).astype(np.int8)
    elif nunique == len(label_keys) + 1 and 0 in np.unique(data):
        logging.debug(
            "Assuming 0 is the background class in label data and hasn't been specified in label_keys."
        )
        label_values = sorted([x for x in np.unique(data).tolist() if x > 0])
        for i, key in zip(label_values, label_keys, strict=True):
            label = np.where((data != i) & (data != -1), 0, data)
            labels[key] = np.where(label == i, 1, label).astype(np.int8)
    else:
        raise ValueError(
            f"Number of unique values in label data ({metadata.nunique}) does not match number of provided label keys ({len(label_keys)})."
        )
    return labels


def load_labels(
    file_path: str | Path, label_keys: list[str], key: str | None
) -> dict[str, np.ndarray]:
    """Load labels from a given file path, given a list of label names in ascending-value order. Supports .h5, .hdf5, .mrc, .mrcs, .tiff, and .tif formats.

    Args:
        file_path: The path to the label file.
        label_keys: A list of label names in ascending-value order (e.g., ['mito', 'cristae'] for 0=background, 1=mito, 2=cristae).
        key: An optional key to specify which dataset to load from an HDF5 file.

    Raises:
        ValueError: If the number of unique values in the label data does not match the number of provided label keys.
        ValueError: If the file format is unsupported.
        ValueError: If the specified key is not found in the label data.
        FileNotFoundError: If the specified file does not exist.

    Returns:
        A dictionary of label name to int8 label array.
    """

    assert (
        key is None or key in label_keys
    ), f"Label key {key} must be one of the specified label keys {label_keys} or None."
    file_path = Path(file_path)
    if not file_path.exists():
        raise FileNotFoundError(f"File {file_path} does not exist.")
    labels = {}
    if file_path.suffix in [".h5", ".hdf", ".hdf5"]:
        _, data, metadata = read_hdf(file_path, key=key)
        if len(label_keys) > 1:
            labels_dict = _match_label_keys_to_data(data, label_keys, metadata)
            labels.update(labels_dict)
        else:
            labels[key] = data.astype(np.int8)
    elif file_path.suffix in [".mrc", ".mrcs"]:
        data, metadata = read_mrc(file_path)
        labels.update(_match_label_keys_to_data(data, label_keys, metadata))
    elif file_path.suffix in [".tiff", ".tif"]:
        data, metadata = read_tiff(file_path)
        labels.update(_match_label_keys_to_data(data, label_keys, metadata))
    else:
        raise ValueError(
            f"Unsupported file format for file {file_path}. Supported formats are .h5, .hdf, .hdf5, .mrc, .mrcs, .tiff, .tif, and image folders."
        )
    return labels


def load_files_from_path(path: Path) -> list[Path]:
    """Load files from a given directory or a .txt file listing file paths.

    Args:
        path (Path): The path to the directory or .txt file.

    Raises:
        ValueError: If the path is not a directory or a .txt file.

    Returns:
        list[Path]: A list of file paths.
    """

    if path.is_dir():
        file_paths = sorted(
            [f for f in path.rglob("*") if f.suffix in tomogram_exts]
        )
    elif path.is_file() and path.suffix == ".txt":
        with open(path) as f:
            file_paths = [Path(line.strip()) for line in f if line.strip()]
    else:
        raise ValueError(
            "Data path must be a directory or a .txt file listing data files."
        )
    assert len(file_paths) > 0, f"No valid tomogram files found in {path}."
    return file_paths


#### Creation Utilities ####


@dataclass
class SavedModel:
    """A class to represent a pre-trained model.

    Attributes:
        name: The name of the model.
        model_type: The type of the model, e.g., 'CryoVIT', '3D U-Net'.
        label_key: The label key used for training the model.
        model_cfg: The config dictionary to instantiate the model.
        weights: The saved weights of the model.
    """

    name: str
    model_type: ModelType
    label_key: str
    model_cfg: BaseModel
    weights: dict[str, Any]


def save_model(
    model_name: str,
    label_key: str,
    model: torch.nn.Module,
    model_cfg: BaseModel,
    save_path: str | Path,
) -> None:
    """Save a model to a given path.

    Args:
        model_name: The name of the model.
        label_key: The label key used for training the model.
        model: The model to save.
        model_cfg: The config dictionary to instantiate the model.
        save_path: The path to save the model to.
    """

    weights = model.state_dict()
    model_type = model_cfg.name.lower()
    saved_model = SavedModel(
        name=model_name,
        model_type=ModelType(model_type),
        label_key=label_key,
        model_cfg=model_cfg,
        weights=weights,
    )
    with open(save_path, "wb") as f:
        pickle.dump(saved_model, f)


def save_model_from_weights(
    model_name: str,
    label_key: str,
    model_type: ModelType,
    weights_path: str | Path,
    save_path: str | Path,
    **kwargs,
) -> None:
    """Save a model to a given path from a weights file.

    Args:
        model_name: The name of the model.
        label_key: The label key used for training the model.
        model_type: The type of the model.
        weights_path: The path to the weights file.
        save_path: The path to save the model to.
        **kwargs: Additional keyword arguments to pass to the model config. To access nested config parameters, use double underscores (e.g., `a.b -> a__b`).

    Raises:
        FileNotFoundError: If the weights file does not exist.
    """
    from hydra import compose, initialize

    if not Path(weights_path).exists():
        raise FileNotFoundError(f"Weights file {weights_path} does not exist.")
    weights = torch.load(weights_path)
    # load model from Hydra config
    with initialize(
        version_base="1.2",
        config_path="configs",
    ):
        cfg = compose(
            config_name="infer_model",
            overrides=[f"model={model_type.value}"]
            + [f"model.{k}={v}" for k, v in kwargs.items()],
        )
    model_dir = Path(__file__).parent / "foundation_models"
    if cfg.model._target_ == "cryovit.models.sam2.SAM2":
        # Load SAM2 pre-trained models
        model = create_sam_model_from_weights(cfg.model, model_dir / "SAM2")
    else:
        model = instantiate(cfg.model)
    model.load_state_dict(weights)

    save_model(model_name, label_key, model, cfg.model, save_path)


def load_model(
    model_path: str | Path, load_model: bool = True
) -> tuple[torch.nn.Module | None, ModelType, str, str]:
    """Load a model from a given path.

    Args:
        model_path: The path to the model file.
        load_model: Whether to load the model weights. If False, only returns the model type, name, and label key.

    Raises:
        FileNotFoundError: If the specified file does not exist.

    Returns:
        A tuple of the model (or None if load_model is False), the model type, the model name, and the label key.
    """

    if not Path(model_path).exists():
        raise FileNotFoundError(f"Model file {model_path} does not exist.")
    with open(model_path, "rb") as f:
        saved_model = pickle.load(f)
    if load_model:
        model_dir = Path(__file__).parent / "foundation_models"
        if saved_model.model_cfg._target_ == "cryovit.models.sam2.SAM2":
            # Load SAM2 pre-trained models
            model = create_sam_model_from_weights(
                saved_model.model_cfg, model_dir / "SAM2"
            )
        else:
            model = instantiate(saved_model.model_cfg)
        model.load_state_dict(saved_model.weights)
    else:
        model = None
    return (
        model,
        saved_model.model_type,
        saved_model.name,
        saved_model.label_key,
    )



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "CryoViT"
copyright = "%Y, Vivian Li, Sanket Gupte"  # noqa: A001
author = "Vivian Li, Sanket Gupte"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.doctest",
    "sphinx.ext.autosummary",
    "sphinx.ext.viewcode",
    "sphinx.ext.githubpages",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

## Autodoc settings
autodoc_member_order = "bysource"
autodoc_default_options = {
    "members": True,
    "show-inheritance": True,
}

## Napoleon settings
napoleon_google_docstring = True
napoleon_include_init_with_doc = True

## Autosummary settings
autosummary_generate = True
autosummary_ignore_module_all = False
add_module_names = False


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]
html_css_files = ["custom.css"]
html_theme_options = {"includehidden": True}



import argparse
from pathlib import Path

from cryovit._logging_config import setup_logging
from cryovit.visualization import (
    process_experiment,
    process_fractional_experiment,
    process_multi_experiment,
    process_multi_label_experiment,
    process_samples,
    process_single_experiment,
    process_sparse_experiment,
)

setup_logging("INFO")
import logging  # noqa: E402

model_names = {
    "cryovit": "CryoViT",
    "unet3d": "3D U-Net",
}

experiment_names = {
    "dino_pca": {},
    "segmentations": {
        s_group: {
            m_key: f"single_{s_group.lower()}_{m_key}" for m_key in model_names
        }
        for s_group in ["HD"]
    },
    "single": {
        s_group: {
            f"single_{s_group.lower()}_{m_key}_mito": [m_value, s_group]
            for m_key, m_value in model_names.items()
        }
        for s_group in ["AD", "HD", "RGC", "Algae"]
    },
    "multi": {
        s_group: {
            m_value: {
                f"{s_group[0]}_to_{s_group[1]}_{m_key}_mito": [
                    m_value,
                    "forward",
                ],
                f"{s_group[1]}_to_{s_group[0]}_{m_key}_mito": [
                    m_value,
                    "backward",
                ],
            }
            for m_key, m_value in model_names.items()
        }
        for s_group in [
            ("hd", "healthy"),
            ("old", "young"),
            ("neuron", "fibro_cancer"),
        ]
    },
    "multi_label": {
        f"fractional_{m_key}_{s_group}": [m_value, s_group]
        for m_key, m_value in model_names.items()
        for s_group in [
            "mito",
            "cristae",
            "microtubule",
            "granule",
            "bacteria",
        ]
    },
    "fractional": {
        s_group: {
            f"fractional_{m_key}_{s_group}": [m_value]
            for m_key, m_value in model_names.items()
        }
        for s_group in [
            "mito",
            "cristae",
            "microtubule",
            "granule",
            "bacteria",
        ]
    },
    "sparse": {
        s_group: {
            f"fractional_cryovit_mito_{s_label.lower()}": [
                f"CryoViT with {s_label} Labels"
            ]
            for s_label in ["Sparse", "Dense"]
        }
        for s_group in ["single", "fractional"]
    },
}

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Main function to visualize the results of certain CryoViT experiments."
    )
    parser.add_argument(
        "--exp_dir",
        type=str,
        required=True,
        help="Directory of experiment results",
    )
    parser.add_argument(
        "--result_dir",
        type=str,
        required=True,
        help="Directory to save results",
    )
    parser.add_argument(
        "--exp_type",
        type=str,
        required=True,
        choices=experiment_names.keys(),
        help="Type of experiment to visualize",
    )
    parser.add_argument(
        "--exp_group",
        type=str,
        default=None,
        required=False,
        help="Experiment group to visualize (e.g., 'hd', 'ad', 'rgc'). All options if not specified.",
    )
    parser.add_argument(
        "--labels",
        type=str,
        nargs="+",
        default=None,
        required=False,
        help="Specific labels to visualize for `segmentation` visualization (e.g., `mito`, `cristae`). All options if not specified.",
    )

    args = parser.parse_args()
    exp_dir = Path(args.exp_dir)
    result_dir = Path(args.result_dir)

    # Sanity checking
    assert (
        exp_dir.exists() and exp_dir.is_dir()
    ), "Experiment directory does not exist or is not a directory."
    if args.exp_group is not None and args.exp_type != "dino_pca":
        assert (
            args.exp_group in experiment_names[args.exp_type]
        ), f"Experiment group {args.exp_group} not found in experiment type {args.exp_type}. Available groups: {list(experiment_names[args.exp_type].keys())}"

    exp_group = (
        [args.exp_group]
        if args.exp_group
        else list(experiment_names[args.exp_type].keys())
    )

    exp_names = {}
    for group in exp_group:
        if group in experiment_names[args.exp_type]:
            exp_names[group] = experiment_names[args.exp_type][group]

    match args.exp_type:
        case "dino_pca":
            process_samples(exp_dir, result_dir, sample=args.exp_group)
        case "segmentations":
            for group in exp_group:
                for model in exp_names[group]:
                    process_experiment(
                        exp_dir,
                        result_dir,
                        exp_template=exp_names[group][model],
                        labels=args.labels,
                    )
        case "single":
            process_single_experiment(
                args.exp_type, "all", exp_names, exp_dir, result_dir
            )
        case "multi":
            for group, model_and_names in exp_names.items():
                combined_names = {}
                for names in model_and_names.values():
                    combined_names.update(names)
                process_multi_experiment(
                    args.exp_type, group, combined_names, exp_dir, result_dir
                )
        case "multi_label":
            process_multi_label_experiment(
                args.exp_type, exp_names, exp_dir, result_dir
            )
        case "fractional":
            for label, names in exp_names.items():
                process_fractional_experiment(
                    args.exp_type,
                    label,
                    names,
                    exp_dir,
                    result_dir,
                    plot_points=True,
                )
                process_fractional_experiment(
                    args.exp_type,
                    label,
                    names,
                    exp_dir,
                    result_dir,
                    plot_points=False,
                )
        case "sparse":
            for group, names in exp_names.items():
                process_sparse_experiment(
                    args.exp_type, group, names, exp_dir, result_dir
                )
        case _:
            logging.error("Unknown experiment type: %s", args.exp_type)



"""Utility functions for datamodules."""

import torch
from torch.nn.functional import pad

from cryovit.types import (
    BatchedTomogramData,
    BatchedTomogramMetadata,
    TomogramData,
)


def collate_fn(batch: list[TomogramData]) -> BatchedTomogramData:  # type: ignore
    """Combine multiple tomograms into a single batch with metadata."""

    # Initialize metadata
    unique_samples = {}  # use dictionaries as ordered sets
    unique_names = {}
    tomo_identifiers = torch.empty(len(batch), 2, dtype=torch.long)
    split_id = torch.empty(len(batch), dtype=torch.int)
    use_splits = True

    tomo_sizes = torch.empty(len(batch), dtype=torch.int)
    min_slices = float("inf")
    # Get tomogram sizes
    for tomo_idx, tomo_data in enumerate(batch):
        D = tomo_data.data.shape[-3]
        tomo_sizes[tomo_idx] = D
        min_slices = min(min_slices, D)

    # Initialize data arrays
    max_size = int(tomo_sizes.max().item())
    C, _, Hp, Wp = batch[0].data.size()
    H, W = batch[0].label.size()[-2:]
    tomo_batch = torch.empty(
        len(batch), C, max_size, Hp, Wp, dtype=torch.float
    )
    labels = torch.empty(len(batch), max_size, H, W, dtype=torch.float)
    aux_data = {key: [] for key in batch[0].aux_data}
    # Loop through batch tomograms
    for tomo_idx, tomo_data in enumerate(batch):
        # Get data
        data = tomo_data.data
        label = tomo_data.label
        for key, value in tomo_data.aux_data.items():
            aux_data[key].append(value)

        # Pad data
        D = data.size()[-3]
        if max_size != D:
            padding = max_size - D
            data = pad(
                data, (0, 0, 0, 0, 0, padding), mode="constant", value=0
            )
            label = pad(
                data, (0, 0, 0, 0, 0, padding), mode="constant", value=-1
            )  # ignore padding
        tomo_batch[tomo_idx] = data
        labels[tomo_idx] = label

        # Get metadata
        sample = tomo_data.sample
        name = tomo_data.tomo_name
        split = tomo_data.split_id

        unique_samples[sample] = None
        s_id = list(unique_samples).index(sample)
        unique_names[name] = None
        n_id = list(unique_names).index(name)
        tomo_identifiers[tomo_idx][0] = s_id
        tomo_identifiers[tomo_idx][1] = n_id
        if split is not None and use_splits:
            split_id[tomo_idx] = split
        else:
            use_splits = False

    tomo_batch = tomo_batch.permute(
        (0, 2, 1, 3, 4)
    )  # (B, C, D, H, W) -> (B, D, C, H, W)
    metadata = BatchedTomogramMetadata(
        samples=list(unique_samples),
        tomo_names=list(unique_names),
        unique_id=tomo_identifiers,
        split_id=split_id if use_splits else None,
    )  # type: ignore
    return BatchedTomogramData(
        tomo_batch=tomo_batch,
        tomo_sizes=tomo_sizes,
        labels=labels,
        aux_data=aux_data,
        metadata=metadata,
        min_slices=min_slices,
    )  # type: ignore



"""Dataset class for loading tomograms for CryoViT scripts."""

import logging
from typing import Any

import numpy as np
import torch
import torch.nn.functional as F
from numpy.typing import NDArray
from torch.utils.data import Dataset
from torchvision.transforms import Normalize

from cryovit.config import DINO_PATCH_SIZE
from cryovit.datasets.vit_dataset import (
    IMAGENET_DEFAULT_MEAN,
    IMAGENET_DEFAULT_STD,
)
from cryovit.types import FileData, TomogramData
from cryovit.utils import load_data, load_labels


class FileDataset(Dataset):
    """A dataset class for handling and preprocessing tomographic data for CryoVIT models."""

    def __init__(
        self,
        files: list[FileData],
        input_key: str | None,
        label_key: str | None,
        train: bool = False,
        for_dino: bool = False,
    ) -> None:
        """Creates a new FileDataset object.

        Args:
            files (list[FileData]): A list of FileData objects containing file paths and metadata.
            input_key (Optional[str]): The key in a HDF5 file to access input features.
            label_key (Optional[str]): The key in a HDF5 file to access labels.
            train (bool): Flag to determine if the dataset is for training (enables transformations).
            for_dino (bool): Flag to determine if the dataset is for DINO feature extraction (enables DINO transformations).
        """

        self.files = files
        self.input_key = input_key
        self.label_key = label_key
        self.train = train
        self.for_dino = for_dino

        self.transform = Normalize(IMAGENET_DEFAULT_MEAN, IMAGENET_DEFAULT_STD)
        self._key_cache = {}
        self._printed_resize_warning = False

    def __len__(self) -> int:
        """Returns the total number of tomograms in the dataset."""

        return len(self.files)

    def __getitem__(self, idx: int) -> TomogramData:  # type: ignore
        """Retrieves a single item from the dataset.

        Args:
            idx (int): The index of the item.

        Returns:
            TomogramData: A dataclass containing the loaded data, labels, and metadata.

        Raises:
            IndexError: If index is out of the range of the dataset.
        """

        if idx >= len(self):
            raise IndexError

        file_data = self.files[idx]
        data = self._load_tomogram(file_data)
        aux_data = {}

        if self.for_dino:
            dino_data = self._dino_transform(data["input"])
            return TomogramData(
                sample=file_data.sample,
                tomo_name=file_data.tomo_path.name,
                split_id=None,
                data=dino_data,
                label=torch.zeros(
                    data["input"].shape, dtype=torch.bool
                ),  # dummy label,
                aux_data={"data": data["input"].squeeze(0)},
            )  # type: ignore
        if self.train:
            self._random_crop(data)
        elif not self.train:  # i.e., eval or predict
            # Load the full tomogram as aux_data for visualization
            aux_data = {
                "data": (
                    load_data(file_data.tomo_path, key="data")[0].squeeze(0)
                    if self.input_key != "data"
                    else data["input"].squeeze(0)
                )
            }

        return TomogramData(
            sample=file_data.sample,
            tomo_name=file_data.tomo_path.name,
            split_id=None,
            data=data["input"],
            label=data["label"],
            aux_data=aux_data,
        )  # type: ignore

    def _load_tomogram(self, file_data: FileData) -> dict[str, Any]:
        """Loads a single tomogram based on the file information.

        Args:
            file_data (FileData): An object containing the file paths and metadata.

        Returns:
            data (dict[str, Any]): A dictionary with input data, label, and any auxiliary data.
        """

        tomo_path = file_data.tomo_path
        label_path = file_data.label_path

        # Cache the label key used for each tomogram to avoid redundant reads
        if tomo_path in self._key_cache:
            data, _ = load_data(tomo_path, key=self._key_cache[tomo_path])
        else:
            data, key = load_data(tomo_path, key=self.input_key)
            self._key_cache[tomo_path] = key
        labels = (
            load_labels(
                label_path, label_keys=file_data.labels, key=self.label_key
            )
            if label_path is not None and file_data.labels is not None
            else None
        )
        assert data is not None, f"Failed to load data from {tomo_path}"

        data_dict = {
            "input": data,
            "label": (
                labels[self.label_key]
                if labels is not None and self.label_key is not None
                else np.zeros(
                    (1, *data.shape[1:]), dtype=np.int8
                )  # replace channel
            ),
        }
        return data_dict

    def _random_crop(self, data: dict[str, Any]) -> None:
        """Applies a random crop to the input data in the record dictionary.

        Args:
            data (dict[str, Any]): The record dictionary containing 'input' and 'label' data.
        """

        max_depth = 128
        side = 32 if self.input_key == "dino_features" else 512
        d, h, w = data["input"].shape[-3:]
        x, y, z = min(d, max_depth), side, side

        if (d, h, w) == (x, y, z):
            return  # nothing to be done

        delta_d = d - x + 1
        delta_h = h - y + 1
        delta_w = w - z + 1

        di = np.random.choice(delta_d) if delta_d > 0 else 0
        hi = np.random.choice(delta_h) if delta_h > 0 else 0
        wi = np.random.choice(delta_w) if delta_w > 0 else 0

        data["input"] = data["input"][
            ..., di : di + x, hi : hi + y, wi : wi + z
        ]

        if self.input_key == "dino_features":
            hi, wi, y, z = 16 * np.array([hi, wi, y, z])

        data["label"] = data["label"][di : di + x, hi : hi + y, wi : wi + z]

    def _dino_transform(self, data: NDArray[np.float32]) -> torch.Tensor:
        """Applies normalization and resizing transformations to the tomogram.

        Args:
            data (NDArray[np.float32]): The loaded tomogram data as a numpy array.

        Returns:
            torch.Tensor: The transformed data as a PyTorch tensor.
        """

        scale = (DINO_PATCH_SIZE / 16, DINO_PATCH_SIZE / 16)
        h, w = data.shape[-2:]
        # Resize height and width to be multiples of 16
        H = int(np.ceil(h / 16) * 16)
        W = int(np.ceil(w / 16) * 16)
        if h != H or w != W:
            if not self._printed_resize_warning:
                logging.warning(
                    "Resizing tomogram from %s to %s", (h, w), (H, W)
                )
                self._printed_resize_warning = True
            data = np.pad(
                data, ((0, 0), (0, 0), (0, H - h), (0, W - w)), mode="edge"
            )
            h, w = H, W
        assert (
            h % 16 == 0 and w % 16 == 0
        ), f"Invalid height: {h} or width: {w}"

        np_data = data.transpose((1, 0, 2, 3))  # D, C, H, W
        np_data = np.repeat(np_data, 3, axis=1)  # C, D, H, W

        torch_data = torch.from_numpy(
            np_data
        ).float()  # data expected to be float32, [0-1]
        torch_data: torch.Tensor = self.transform(torch_data)[
            :, [0], :, :
        ]  # D, C, H, W
        torch_data: torch.Tensor = F.interpolate(
            torch_data, scale_factor=scale, mode="bicubic"
        )
        return torch_data



"""Custom types and dataclasses for CryoViT models."""

from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Any

import numpy as np
import torch
from numpy.typing import NDArray
from tensordict import tensorclass


#### Enum Definitions ####
class Sample(Enum):
    """Enum of all valid CryoET Samples."""

    BACHD = "BACHD"
    BACHD_Microtubules = "BACHD Microtubules"
    dN17_BACHD = "dN17 BACHD"
    Q109 = "Q109"
    Q109_Microtubules = "Q109 Microtubules"
    Q18 = "Q18"
    Q18_Microtubules = "Q18 Microtubules"
    Q20 = "Q20"
    Q53 = "Q53"
    Q53_KD = "Q53 PIAS1"
    Q66 = "Q66"
    Q66_GRFS1 = "Q66 GRFS1"
    Q66_KD = "Q66 PIAS1"
    WT = "Wild Type"
    WT_Microtubules = "Wild Type Microtubules"
    cancer = "Cancer"
    AD = "AD"
    AD_Abeta = "AD Abeta"
    Aged = "Aged"
    Young = "Young"
    RGC_CM = "RGC CM"
    RGC_control = "RGC Control"
    RGC_naPP = "RGC naPP"
    RGC_PP = "RGC PP"
    CZI_Algae = "Algae"
    CZI_Campy_C = "Campy C"
    CZI_Campy_CDel = "Campy C-Deletion"
    CZI_Campy_F = "Campy F"
    CZI_Fibroblast = "Mouse Fibroblast"


class ModelType(Enum):
    """Enum of all valid model types."""

    CRYOVIT = "cryovit"
    UNET3D = "unet3d"
    SAM2 = "sam2"
    MEDSAM = "medsam"


#### Class Definitions ####


@dataclass
class FileData:
    """
    This class represents the file data for a single tomogram.

    Attributes:
        tomo_path: A path to the raw tomogram data.
        label_path: A path to the segmentation labels. None if not available.
        labels: A list of strings representing the label names. None if not available.
        sample: A string representing the sample. None if not available.
    """

    tomo_path: Path
    label_path: Path | None = None
    labels: list[str] | None = None
    sample: str | None = None


@tensorclass  # type: ignore
class TomogramData:
    """
    This class represents the data in single tomogram.

    Attributes:
        sample: A string representing the experiment sample.
        tomo_name: A string for the file_path of the raw tomogram data.
        split_id: An optional identifier for training/val/test splits.
        data: A [DxCxHxW] tensor containing the tomogram data.
        label: A [DxHxW] tensor containing the segmentation labels.
        aux_data: An optional dictionary containing additional data, such as raw data input for dino_features.
    """

    sample: str
    tomo_name: str
    split_id: int | None

    data: torch.FloatTensor
    label: torch.BoolTensor
    aux_data: dict[str, Any] | None = None


@tensorclass  # type: ignore
class BatchedTomogramMetadata:
    """
    This class represents metadata about a batch of tomograms.

    Attributes:
        samples: A list containing all possible samples of tomogram files in the batch.
        tomo_names: A list containing all possible names of tomogram files in the batch.
        unique_id: A [Bx2] tensor containing the corresponding index in samples and tomo_names for each tomogram. Index consists of (sample_id, tomo_name_id).
        split_id: An optional list containing the training/val/test split the tomogram is a part of.
    """

    samples: list[str]
    tomo_names: list[str]
    unique_id: torch.LongTensor
    split_id: list[torch.IntTensor] | None

    @property
    def identifiers(self) -> tuple[list[str], list[str]]:
        samples = [self.samples[i[0].item()] for i in self.unique_id]  # type: ignore
        names = [self.tomo_names[i[1].item()] for i in self.unique_id]  # type: ignore
        return samples, names


@tensorclass  # type: ignore
class BatchedTomogramData:
    """
    This class represents a batch of tomograms with associated annotations.

    Attributes:
        tomo_batch: A [[BxDxCxHxW] tensor containing the tomogram data for each slice in the batch, where D is a tomogram's depth, and B is the number of tomograms in the batch. The D dimension is padded to the max in the batch.
        tomo_sizes: A [B] tensor containing the size (D) of each tomogram in the batch.
        labels: A [[BxDxHxW] tensor containing the binary labels for segmentation objects in the batch.
        metadata: An instance of BatchedTomogramMetadata containing metadata about the batch and the tomograms inside.
        min_slices: An integer representing the minimum number of slices of all tomograms in the batch.
        aux_data: A dictionary containing additional data as a list of values, such as raw data input for dino_features.
    """

    tomo_batch: torch.Tensor
    tomo_sizes: torch.Tensor
    labels: torch.Tensor
    metadata: BatchedTomogramMetadata  # type: ignore
    min_slices: int
    aux_data: dict[str, list[Any]] | None = None

    def pin_memory(self, device=None):
        self.tomo_batch = self.tomo_batch.pin_memory(device=device)
        self.tomo_sizes = self.tomo_sizes.pin_memory(device=device)
        self.labels = self.labels.pin_memory(device=device)
        return self

    @property
    def num_tomos(self) -> int:
        """Returns the number of tomograms in the batch."""

        return self.tomo_batch.shape[0]

    @property
    def num_slices(self) -> int:
        """Returns the maximum number of slices in the batch."""

        return self.tomo_batch.shape[1]

    def index_to_flat_batch(self, idx: int) -> torch.Tensor:
        """Returns a [BxD] tensor containing the indices corresponding to a certain slice in a flat batch tensor."""

        if idx >= self.num_slices:
            raise IndexError(
                f"Slice index {idx} is out of bounds for the maximum number of slices {self.num_slices}."
            )
        batch_idxs = torch.argwhere(self.tomo_sizes > idx).long()
        batch_sizes = self.tomo_sizes[batch_idxs].flatten()
        batch_ll = torch.cumsum(batch_sizes, dim=0) - batch_sizes
        slice_idxs = batch_ll + idx
        return slice_idxs.long()

    @property
    def flat_tomo_batch(self) -> torch.Tensor:
        """Returns a [[BxD]xCxHxW] tensor from a [BxDxCxHxW] tensor (C is optional)."""

        return self.tomo_batch.reshape(-1, *self.tomo_batch.shape[2:])


@dataclass
class BatchedModelResult:
    """
    This class represents the model result from a batch of tomograms, organized per tomogram.

    Attributes:
        num_tomos: The number of tomograms in the batch.
        samples: The sample for each tomogram in the batch.
        tomo_names: The file name for each tomogram in the batch.
        split_id: The optional split id for each tomogram in the batch.
        data: The raw tomogram data for each tomogram in the batch.
        label: The true segmentation labels for each tomogram in the batch.
        preds: The model predictions for each tomogram in the batch.
        losses: A dictionary of losses for each tomogram in the batch.
        metrics: A dictionary of metrics for each tomogram in the batch.
        aux_data: An optional dictionary containing auxiliary data for each tomogram in the batch.
    """

    num_tomos: int
    samples: list[str]
    tomo_names: list[str]
    split_id: list[int] | None
    data: list[NDArray[np.float32]]
    label: list[NDArray[np.uint8]]
    preds: list[NDArray[np.float32]]
    losses: dict[str, float]
    metrics: dict[str, float]
    aux_data: dict[str, list[Any]] | None = None



"""Script for setting up and training CryoVIT models based on configuration files."""

import logging
from collections.abc import Iterable
from pathlib import Path

import torch
from hydra import compose, initialize
from hydra.core.hydra_config import HydraConfig
from hydra.utils import instantiate
from pytorch_lightning import Trainer, seed_everything
from pytorch_lightning.loggers import TensorBoardLogger

from cryovit.config import BaseExperimentConfig
from cryovit.models import create_sam_model_from_weights
from cryovit.types import ModelType
from cryovit.utils import load_model, save_model

torch.set_float32_matmul_precision("high")

## For Scripts


def run_training(
    train_data: list[Path],
    train_labels: list[Path],
    labels: list[str],
    model_type: ModelType,
    model_name: str,
    label_key: str,
    result_dir: Path,
    val_data: list[Path] | None = None,
    val_labels: list[Path] | None = None,
    num_epochs: int = 50,
    log_training: bool = False,
    ckpt_path: Path | None = None,
) -> Path:
    """Run training on the specified data and labels.

    Args:
        train_data (list[Path]): List of paths to the training tomograms.
        train_labels (list[Path]): List of paths to the training labels.
        labels (list[str]): List of label names to train on.
        model_type (ModelType): Type of the model to train.
        model_name (str): Name of the model.
        label_key (str): Key for the label in the dataset.
        result_dir (Path): Directory where the training results will be saved.
        val_data (Optional[list[Path]], optional): List of paths to the validation tomograms. Defaults to None.
        val_labels (Optional[list[Path]], optional): List of paths to the validation labels. Defaults to None.
        num_epochs (int, optional): Number of training epochs. Defaults to 50.
        log_training (bool, optional): Whether to log training metrics to Tensorboard. Defaults to False.
        ckpt_path (Optional[Path], optional): Path to a .model file, or .ckpt/.pt file to fine-tune from. Defaults to None.

    Returns:
        Path: Path to the saved model file.
    """

    ## Setup hydra config
    with initialize(
        version_base="1.2",
        config_path="../configs",
        job_name="cryovit_train",
    ):
        cfg = compose(
            config_name="train_model",
            overrides=[
                f"name={model_name}",
                f"label_key={label_key}",
                f"model={model_type.value}",
                "datamodule=file",
                f"trainer.max_epochs={num_epochs}",
            ],
        )
    cfg.paths.model_dir = Path(__file__).parent.parent / "foundation_models"
    save_model_path = result_dir / f"{model_name}.model"

    # Check input key
    if cfg.model.input_key != "dino_features":
        cfg.model.input_key = None  # find available data instead

    ## Setup dataset
    dataset_fn = instantiate(cfg.datamodule.dataset)
    dataloader_fn = instantiate(cfg.datamodule.dataloader)
    datamodule = instantiate(cfg.datamodule, _convert_="all")(
        data_paths=train_data,
        data_labels=train_labels,
        labels=labels,
        val_paths=val_data,
        val_labels=val_labels,
        dataloader_fn=dataloader_fn,
        dataset_fn=dataset_fn,
    )
    logging.info("Setup dataset.")

    ## Setup training
    callbacks = [instantiate(cb_cfg) for cb_cfg in cfg.callbacks.values()]
    loggers = []
    if log_training:
        # tensorboard logger to avoid wandb account issues
        loggers.append(TensorBoardLogger(save_dir=result_dir, name=model_name))
        logging.info(
            "Setup TensorBoard logger. View logs with `tensorboard --logdir %s`",
            result_dir / model_name,
        )
    trainer = instantiate(cfg.trainer, callbacks=callbacks, logger=loggers)

    if ckpt_path is not None and ckpt_path.suffix == ".model":
        # Fine-tune from .model file
        model, _, _, _ = load_model(ckpt_path, load_model=True)
        assert model is not None
    else:
        # Create new model
        if cfg.model._target_ == "cryovit.models.sam2.SAM2":
            # Load SAM2 pre-trained models
            model = create_sam_model_from_weights(
                cfg.model, cfg.paths.model_dir / cfg.paths.sam_name
            )
        else:
            model = instantiate(cfg.model)

        # Optionally load weights from checkpoint
        if ckpt_path is not None:
            if ckpt_path.suffix == ".pt":
                model.load_state_dict(torch.load(ckpt_path))
            elif ckpt_path.suffix == ".ckpt":
                model = model.load_from_checkpoint(ckpt_path)
            else:
                raise ValueError(
                    f"Unsupported checkpoint format: {ckpt_path.suffix}. Use .pt or .ckpt files."
                )
    logging.info("Loaded model.")

    # Base SAM2 only supports image encoder compilation
    if cfg.model._target_ == "cryovit.models.sam2.SAM2":
        logging.info("Compiling image encoder for SAM2 model.")
        try:
            model.compile()
        except Exception as e:  # noqa: BLE001
            logging.error("Unable to compile image encoder for SAM2: %s", e)
    else:
        logging.info("Compiling model forward pass.")
        try:
            model.forward = torch.compile(model.forward)
        except Exception as e:  # noqa: BLE001
            logging.error("Unable to compile forward pass: %s", e)

    logging.info("Starting training.")
    trainer.fit(model, datamodule=datamodule)

    # Save model
    logging.info("Saving model.")
    save_model(model_name, label_key, model, cfg.model, save_model_path)
    return save_model_path


## For Experiments


def setup_exp_dir(cfg: BaseExperimentConfig) -> BaseExperimentConfig:
    """Setup the experiment directory structure and optionally, the W&B logger."""
    # Convert paths to Paths
    cfg.paths.model_dir = Path(cfg.paths.model_dir)
    cfg.paths.data_dir = Path(cfg.paths.data_dir)
    cfg.paths.exp_dir = Path(cfg.paths.exp_dir)
    cfg.paths.results_dir = Path(cfg.paths.results_dir)

    if not isinstance(cfg.datamodule.sample, str) and isinstance(
        cfg.datamodule.sample, Iterable
    ):
        sample = "_".join(sorted(cfg.datamodule.sample))
    else:
        sample = cfg.datamodule.sample
    if not isinstance(cfg.datamodule.test_sample, str) and isinstance(
        cfg.datamodule.test_sample, Iterable
    ):
        test_sample = "_".join(sorted(cfg.datamodule.test_sample))
    else:
        test_sample = cfg.datamodule.test_sample

    new_exp_dir = cfg.paths.exp_dir / cfg.name / sample
    new_exp_dir.mkdir(parents=True, exist_ok=True)
    if cfg.datamodule.split_id is not None:
        new_exp_dir = new_exp_dir / f"split_{cfg.datamodule.split_id}"
    if "Fractional" in cfg.datamodule._target_ and test_sample is not None:
        new_exp_dir.mkdir(parents=True, exist_ok=True)
        new_exp_dir = new_exp_dir / f"test_{test_sample}"

    new_exp_dir.mkdir(parents=True, exist_ok=True)
    cfg.paths.exp_dir = new_exp_dir

    # Setup WandB Logger
    for name, lg in cfg.logger.items():
        if name == "wandb":
            lg_name = (
                str(test_sample) if test_sample is not None else str(sample)
            )
            lg.name = (
                f"{lg_name}_{cfg.datamodule.split_id}"
                if cfg.datamodule.split_id is not None
                else lg_name
            )

    return cfg


def run_trainer(cfg: BaseExperimentConfig) -> None:
    """Sets up and runs the training process using the specified configuration.

    Args:
        cfg (TrainModelConfig): Configuration object containing all settings for the training process.
    """

    seed_everything(cfg.random_seed, workers=True)

    # Setup experiment directories
    cfg = setup_exp_dir(cfg)
    ckpt_path = (
        cfg.paths.exp_dir / "last.ckpt"
        if cfg.ckpt_path is None
        else cfg.ckpt_path
    )
    weights_path = cfg.paths.exp_dir / "weights.pt"

    # Setup dataset
    dataset_fn = instantiate(cfg.datamodule.dataset)
    dataloader_fn = instantiate(cfg.datamodule.dataloader)
    split_file = cfg.paths.data_dir / cfg.paths.csv_name / cfg.paths.split_name
    datamodule = instantiate(cfg.datamodule, _convert_="all")(
        split_file=split_file,
        dataloader_fn=dataloader_fn,
        dataset_fn=dataset_fn,
    )
    logging.info("Setup dataset.")

    # Setup training
    callbacks = [instantiate(cb_cfg) for cb_cfg in cfg.callbacks.values()]
    loggers = [instantiate(lg_cfg) for lg_cfg in cfg.logger.values()]
    trainer: Trainer = instantiate(
        cfg.trainer, callbacks=callbacks, logger=loggers
    )
    logging.info("Setup trainer.")
    if cfg.model._target_ == "cryovit.models.sam2.SAM2":
        # Load SAM2 pre-trained models
        model = create_sam_model_from_weights(
            cfg.model, cfg.paths.model_dir / cfg.paths.sam_name
        )
    else:
        model = instantiate(cfg.model)
    logging.info("Setup model.")

    # Log hyperparameters
    if trainer.loggers:
        hparams = {
            "datamodule_type": HydraConfig.get().runtime.choices["datamodule"],
            "model_name": cfg.model.name,
            "label_key": cfg.label_key,
            "experiment": cfg.name,
            "split_id": cfg.datamodule.split_id,
            "sample": (
                "_".join(sorted(cfg.datamodule.sample))
                if isinstance(cfg.datamodule.sample, Iterable)
                else cfg.datamodule.sample
            ),
            "test_sample": cfg.datamodule.test_sample,
            "cfg": cfg,
            "model": model,
            "model/params/total": sum(p.numel() for p in model.parameters()),
            "model/params/trainable": sum(
                p.numel() for p in model.parameters() if p.requires_grad
            ),
            "model/params/non_trainable": sum(
                p.numel() for p in model.parameters() if not p.requires_grad
            ),
            "datamodule": datamodule,
            "trainer": trainer,
            "resume_ckpt": cfg.resume_ckpt,
            "ckpt_path": cfg.ckpt_path,
            "seed": cfg.random_seed,
        }
        if cfg.model._target_ == "cryovit.models.sam2.SAM2":
            hparams["prompt_lr"] = (
                cfg.model.custom_kwargs.get("prompt_lr", None)
                if cfg.model.custom_kwargs
                else None
            )
        for lg in trainer.loggers:
            lg.log_hyperparams(hparams)

    # Base SAM2 only supports image encoder compilation
    if cfg.model._target_ == "cryovit.models.sam2.SAM2":
        logging.info("Compiling image encoder for SAM2 model.")
        try:
            model.compile()
        except Exception as e:  # noqa: BLE001
            logging.warning("Unable to compile image encoder for SAM2: %s", e)
    else:
        logging.info("Compiling model forward pass.")
        try:
            model.forward = torch.compile(model.forward)
        except Exception as e:  # noqa: BLE001
            logging.warning("Unable to compile forward pass: %s", e)

    logging.info("Starting training.")
    if cfg.resume_ckpt and ckpt_path.exists():
        logging.info("Resuming training from checkpoint: %s", ckpt_path)
        trainer.fit(model, datamodule=datamodule, ckpt_path=str(ckpt_path))
    else:
        trainer.fit(model, datamodule=datamodule)

    # Save model
    logging.info("Saving model.")
    torch.save(model.state_dict(), weights_path)



"""Functions to load data and run DINOv2 feature extraction."""

from pathlib import Path

import h5py
import numpy as np
import pandas as pd
import torch
from hydra import compose, initialize
from hydra.utils import instantiate
from numpy.typing import NDArray
from rich.progress import track

from cryovit.config import (
    DEFAULT_WINDOW_SIZE,
    BaseDataModule,
    DinoFeaturesConfig,
    samples,
    tomogram_exts,
)
from cryovit.types import FileData
from cryovit.visualization.dino_pca import export_pca

torch.set_float32_matmul_precision("high")  # ensures tensor cores are used
dino_model = (
    "facebookresearch/dinov2",
    "dinov2_vitg14_reg",
)  # the giant variant of DINOv2


@torch.inference_mode()
def _dino_features(
    data: torch.Tensor,
    model: torch.nn.Module,
    batch_size: int,
    window_size: tuple[int, int] | int | None = DEFAULT_WINDOW_SIZE,
) -> NDArray[np.float16]:
    """Extract patch features from a tomogram using a DINOv2 model. Uses windows to fit everything in memory.

    Args:
        data (torch.Tensor): The input data tensors containing the tomogram's data.
        model (nn.Module): The pre-loaded DINOv2 model used for feature extraction.
        batch_size (int): The maximum number of 2D slices of a tomograms processed in each batch.
        window_size (Optional[tuple[int, int] | int]): The size of the windows to use for feature extraction.

    Returns:
        NDArray[np.float16]: A numpy array containing the extracted features in reduced precision.
    """
    data = data.cuda()
    w, h = (
        np.array(data.shape[-2:]) // 14
    )  # the number of patches extracted per slice
    all_features = []

    for i in range(0, len(data), batch_size):
        # Check for overflows
        if i + batch_size > len(data):
            batch_size = len(data) - i
        vec = data[i : i + batch_size]
        features = model.forward_features(vec)["x_norm_patchtokens"]  # type: ignore
        features = features.reshape(features.shape[0], w, h, -1)
        features = features.permute([3, 0, 1, 2]).contiguous()
        features = features.to("cpu").half().numpy()
        all_features.append(features)

    return np.concatenate(all_features, axis=1)  # type: ignore


def _save_data(
    data: dict[str, NDArray[np.uint8]],
    features: NDArray[np.float16],
    tomo_name: str,
    dst_dir: Path,
) -> None:
    """Save extracted features to a specified directory as an .hdf file, copying the source tomogram data."""

    dst_dir.mkdir(parents=True, exist_ok=True)

    with h5py.File(dst_dir / tomo_name, "w") as fh:
        for key in data:
            if key != "data":
                if "labels" not in fh:
                    fh.create_group("labels")
                fh["labels"].create_dataset(key, data=data[key], shape=data[key].shape, dtype=data[key].dtype, compression="gzip")  # type: ignore
            else:
                fh.create_dataset(
                    "data",
                    data=data[key],
                    shape=data[key].shape,
                    dtype=data[key].dtype,
                    compression="gzip",
                )
        fh.create_dataset(
            "dino_features",
            data=features,
            shape=features.shape,
            dtype=features.dtype,
        )


def _process_sample(
    src_dir: Path,
    dst_dir: Path,
    csv_dir: Path,
    model: torch.nn.Module,
    sample: str,
    datamodule: BaseDataModule,
    batch_size: int,
    image_dir: Path | None,
):
    """Process all tomograms in a single sample by extracting and saving their DINOv2 features."""

    tomo_dir = src_dir / sample
    result_dir = dst_dir / sample
    csv_file = csv_dir / f"{sample}.csv"
    # If no .csv file, use all tomograms in the dataset
    if not csv_file.exists():
        records = [
            f.name for f in tomo_dir.glob("*") if f.suffix in tomogram_exts
        ]
    else:
        records = pd.read_csv(csv_file)["tomo_name"].to_list()

    dataset = instantiate(datamodule.dataset, data_root=tomo_dir)(
        records=records
    )
    dataloader = instantiate(datamodule.dataloader)(dataset=dataset)

    for i, x in track(
        enumerate(dataloader),
        description=f"[green]Computing features for {sample}",
        total=len(dataloader),
    ):
        features = _dino_features(x, model, batch_size)

        data = {}
        with h5py.File(tomo_dir / records[i]) as fh:
            for key in fh:
                data[key] = fh[key][()]  # type: ignore

        _save_data(data, features, records[i], result_dir)
        # Save PCA calculation of features
        if image_dir is not None:
            export_pca(data["data"], features.astype(np.float32), records[i][:-4], image_dir / sample)  # type: ignore


## For Scripts


def run_dino(
    train_data: list[Path],
    result_dir: Path,
    batch_size: int,
    window_size: int | None = DEFAULT_WINDOW_SIZE,
    visualize: bool = False,
) -> None:
    """Run DINO feature extraction on the specified training data, and saves the results as .hdf files.
    The saved result file will contain `data`, `dino_features`, and any labels present in the source tomogram in the `labels/` group.

    Args:
        train_data (list[Path]): List of paths to the training tomograms.
        result_dir (Path): Directory where the results will be saved.
        batch_size (int): Number of samples to process in each batch.
        window_size (Optional[int], optional): Size of the sliding window for feature extraction. If None, uses the default size.
        visualize (bool, optional): Whether to visualize the extracted features. Defaults to False.
    """

    ## Setup hydra config
    with initialize(
        version_base="1.2",
        config_path="../configs",
        job_name="cryovit_features",
    ):
        cfg = compose(
            config_name="dino_features",
            overrides=[
                f"batch_size={batch_size}",
                "sample=null",
                "export_features=False",
                "datamodule/dataset=file",
            ],
        )
    cfg.paths.model_dir = Path(__file__).parent.parent / "foundation_models"

    ## Load DINOv2 model
    torch.hub.set_dir(cfg.dino_dir)
    model = torch.hub.load(*dino_model, verbose=False).cuda()  # type: ignore
    model.eval()

    ## Setup dataset
    assert (
        len(train_data) > 0
    ), "No valid tomogram files found in the specified training data path."
    train_file_datas = [FileData(tomo_path=f) for f in train_data]
    dataset = instantiate(
        cfg.datamodule.dataset, input_key=None, label_key=None
    )(train_file_datas, for_dino=True)
    dataloader = instantiate(cfg.datamodule.dataloader)(dataset=dataset)

    result_list = [result_dir / f"{f.stem}.hdf" for f in train_data]

    ## Iterate through dataloader and extract features
    try:
        for i, x in track(
            enumerate(dataloader),
            description="[green]Computing DINO features for training data",
            total=len(dataloader),
        ):
            features = _dino_features(
                x.data, model, cfg.batch_size, window_size=window_size
            )

            result_path = result_list[i].with_suffix(".hdf")
            result_path.parent.mkdir(parents=True, exist_ok=True)
            result_dir, tomo_name = result_path.parent, result_path.name
            _save_data(x.aux_data, features, tomo_name, result_dir)
            if visualize:
                export_pca(
                    x.aux_data["data"],
                    features,
                    tomo_name[:-4],
                    result_dir.parent / "dino_images" / tomo_name[:-4],
                )
    except torch.OutOfMemoryError:
        print(
            f"Ran out of GPU memory during DINO feature extraction. Try reducing the batch size or window size. Current batch size is {cfg.batch_size} and window size is {window_size}."
        )
        return


## For Experiments


def run_trainer(cfg: DinoFeaturesConfig) -> None:
    """Sets up and executes the DINO feature computation using the specified configuration.

    Args:
        cfg (DinoFeaturesConfig): Configuration object containing all settings for the DINO feature computation.
    """

    # Convert paths to Paths
    cfg.paths.model_dir = Path(cfg.paths.model_dir)
    cfg.paths.data_dir = Path(cfg.paths.data_dir)
    cfg.paths.exp_dir = Path(cfg.paths.exp_dir)

    src_dir = cfg.paths.data_dir / cfg.paths.tomo_name
    dst_dir = cfg.paths.data_dir / cfg.paths.feature_name
    csv_dir = cfg.paths.data_dir / cfg.paths.csv_name
    image_dir = cfg.paths.exp_dir / "dino_images"
    sample_names = (
        [cfg.sample.name]
        if cfg.sample is not None
        else [s for s in samples if (src_dir / s).exists()]
    )

    torch.hub.set_dir(cfg.dino_dir)
    model = torch.hub.load(*dino_model, verbose=False).cuda()  # type: ignore
    model.eval()

    for sample_name in sample_names:
        _process_sample(
            src_dir,
            dst_dir,
            csv_dir,
            model,
            sample_name,
            cfg.datamodule,  # type: ignore
            cfg.batch_size,
            image_dir if cfg.export_features else None,
        )



import traceback
import warnings

import hydra

from cryovit._logging_config import setup_logging
from cryovit.config import DinoFeaturesConfig, validate_dino_config
from cryovit.run import dino_features

setup_logging("DEBUG")
import logging  # noqa: E402

warnings.simplefilter("ignore")


@hydra.main(
    config_path="../configs",
    config_name="dino_features",
    version_base="1.2",
)
def main(cfg: DinoFeaturesConfig) -> None:
    """Main function to process DINOv2 feature extraction.

    Validates the configuration and processes the sample as per the specified settings. Errors during
    processing are caught and logged.

    Args:
        cfg (DinoFeaturesConfig): Configuration object loaded from dino_features.yaml.
    """

    validate_dino_config(cfg)

    try:
        dino_features.run_trainer(cfg)
    except BaseException as err:  # noqa: BLE001
        logging.error("%s: %s", type(err).__name__, err)
        logging.error(traceback.format_exc())


if __name__ == "__main__":
    main()



"""Implementations of PyTorch Lightning DataModules for loading Cryo-EM tomograms."""

from cryovit.datamodules.base_datamodule import BaseDataModule
from cryovit.datamodules.file_datamodule import FileDataModule
from cryovit.datamodules.fractional_datamodule import FractionalDataModule
from cryovit.datamodules.fractional_sample_datamodule import (
    FractionalSampleDataModule,
)
from cryovit.datamodules.multi_sample_datamodule import MultiSampleDataModule
from cryovit.datamodules.single_sample_datamodule import SingleSampleDataModule

__all__ = [
    "BaseDataModule",
    "FractionalDataModule",
    "FractionalSampleDataModule",
    "SingleSampleDataModule",
    "MultiSampleDataModule",
    "FileDataModule",
]



"""Implementation of the single sample data module."""

import pandas as pd

from cryovit.datamodules.base_datamodule import BaseDataModule


class SingleSampleDataModule(BaseDataModule):
    """Data module for CryoVIT experiments involving a single sample."""

    def __init__(
        self,
        sample: list[str],
        split_id: int | None,
        split_key: str,
        test_sample: list[str] | None = None,
        **kwargs,
    ) -> None:
        """Create a datamodule for training and testing on a single sample.

        Args:
            sample (list[str]): The sample to train on.
            split_id (Optional[int]): An optional split_id to validate with.
            split_key (str): The key used to select splits using split_id.
            test_sample (Optional[list[str]]): The sample to test on. If None, test on the validation set.
        """

        super().__init__(**kwargs)
        # Validity checks
        assert (
            len(sample) == 1
        ), f"Single sample 'sample' should be a single string list. Got {sample} instead."
        assert (
            test_sample is None or len(test_sample) == 1
        ), f"Single sample 'test_sample' should be a single string list or None. Got {test_sample} instead."

        self.sample = sample[0]
        self.split_id = split_id
        self.split_key = split_key
        self.test_sample = (
            test_sample[0] if test_sample is not None else test_sample
        )

    def train_df(self) -> pd.DataFrame:
        """Train tomograms: exclude those from the sample with the specified split_id.

        Returns:
            pd.DataFrame: A dataframe specifying the train tomograms.
        """

        assert self.record_df is not None
        if self.split_id is not None:
            df = self.record_df[
                (self.record_df[self.split_key] != self.split_id)
                & (self.record_df["sample"] == self.sample)
            ]
        else:
            df = self.record_df[self.record_df["sample"] == self.sample][
                ["sample", "tomo_name"]
            ]
        return df

    def val_df(self) -> pd.DataFrame:
        """Validation tomograms: optionally validate on tomograms with the specified split_id.

        Returns:
            pd.DataFrame: A dataframe specifying the validation tomograms.
        """

        assert self.record_df is not None
        if self.split_id is None:  # validate on train set
            return self.train_df()

        return self.record_df[
            (self.record_df[self.split_key] == self.split_id)
            & (self.record_df["sample"] == self.sample)
        ]

    def test_df(self) -> pd.DataFrame:
        """Test tomograms: test on tomograms from the specified test_sample or split_id.

        Returns:
            pd.DataFrame: A dataframe specifying the test tomograms.
        """

        assert self.record_df is not None
        if self.test_sample is None:
            return self.val_df()

        # If testing on another sample, use the whole sample
        return self.record_df[self.record_df["sample"] == self.test_sample][
            ["sample", "tomo_name"]
        ]

    def predict_df(self) -> pd.DataFrame:
        """Predict tomograms: predict on the whole sample.

        Returns:
            pd.DataFrame: A dataframe specifying the predict tomograms.
        """

        assert self.record_df is not None
        return self.record_df[self.record_df["sample"] == self.sample][
            ["sample", "tomo_name"]
        ]



"""Functions to run feature extraction, training, evaluation, and inference for users."""

from .dino_features import run_dino
from .eval_model import run_evaluation
from .infer_model import run_inference
from .train_model import run_training

__all__ = ["run_dino", "run_evaluation", "run_inference", "run_training"]



"""3D U-Net-based prompt predictor for SAM2 using image encodings and LoRA adaptation modules.

Prompt predictor architecture is based on the prompt predictor in https://github.com/ChengyinLee/AutoProSAM_2024/ and https://github.com/MIC-DKFZ/nnUNet.
"""

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from sam2.modeling.sam.transformer import Attention
from torch import Tensor


class PromptConvBlock(nn.Module):
    """A convolutional block used in the prompt predictor."""

    def __init__(self, in_channels: int, out_channels: int):
        super().__init__()
        self.layers = nn.Sequential(
            nn.Conv3d(
                in_channels, out_channels, kernel_size=3, padding="same"
            ),
            nn.InstanceNorm3d(out_channels, eps=1e-3, affine=True),
            nn.GELU(),
            nn.Conv3d(
                out_channels, out_channels, kernel_size=3, padding="same"
            ),
            nn.InstanceNorm3d(out_channels, eps=1e-3, affine=True),
            nn.GELU(),
        )

    def forward(self, x: Tensor) -> Tensor:
        return self.layers(x)


class PromptAnalysisBlock(nn.Module):
    """A single block in the down-sampling UNet architecture for prompt prediction."""

    def __init__(self, in_channels: int, out_channels: int, scale: int = 2):
        super().__init__()
        self.layers = nn.Sequential(
            nn.MaxPool3d(scale),
            PromptConvBlock(in_channels, out_channels),
        )

    def forward(self, x: Tensor) -> Tensor:
        return self.layers(x)


class PromptSynthesisBlock(nn.Module):
    """A single block in the up-sampling UNet architecture for prompt prediction."""

    def __init__(
        self,
        in_channels: int,
        skip_channels: int,
        out_channels: int,
        scale: int = 2,
    ):
        super().__init__()
        self.upconv = nn.ConvTranspose3d(
            in_channels, in_channels, kernel_size=scale, stride=scale
        )
        self.layers = nn.Sequential(
            PromptConvBlock(in_channels + skip_channels, out_channels),
            PromptConvBlock(out_channels, out_channels),
        )
        self.scale = scale

    def forward(self, x: Tensor, skip_x: Tensor) -> Tensor:
        x = self.upconv(x)
        x = F.interpolate(
            x, size=skip_x.shape[-3:], mode="trilinear", align_corners=True
        )  # handle size mismatches
        x = torch.cat([x, skip_x], dim=1)
        x = self.layers(x)
        return x


class PromptBoxPredictor(nn.Module):
    """A simple box predictor using a linear layer."""

    def __init__(self, in_channels: int):
        super().__init__()
        self.pool = nn.AdaptiveAvgPool2d(1)
        self.fc = nn.Linear(
            in_channels, 4
        )  # 4 channels for box (x1, y1, x2, y2)

    def forward(self, x: Tensor) -> Tensor:
        # Global average pooling
        B, C, D, H, W = x.shape
        x = self.pool(x)  # [B, C, D, 1, 1]
        x = x.transpose(1, 2).view(B * D, -1, 1, 1)  # [B*D, C, 1, 1]
        x = x.flatten(1)  # [B*D, C]
        x = self.fc(x)  # [B*D, 4]
        x = torch.sigmoid(x)  # Normalize to [0, 1]
        x1y1 = x[:, :2]
        x2y2 = x[:, 2:] + x1y1  # Ensure x2y2 >= x1y1
        x = torch.cat([x1y1, x2y2], dim=1)  # [B*D, 4]
        return x


class PromptPredictor(nn.Module):
    """A simple UNet to predict mask prompts for the SAMv2 model from image encodings."""

    def __init__(
        self,
        in_channels: int = 256,
        depth: int = 3,
        layer_scale: int = 2,
        channel_dims: list[int] | None = None,
    ):
        if channel_dims is None:
            channel_dims = [64, 128, 256]
        super().__init__()
        assert depth == len(
            channel_dims
        ), "Depth must match the length of channel_dims"
        self.scale_factor = 4  # Scale factor for mask prediction
        # Original SAM2 has 4x patch embedding

        self.init_conv = PromptConvBlock(in_channels, channel_dims[0])

        self.down_layers = nn.ModuleList(
            [
                PromptAnalysisBlock(
                    channel_dims[i], channel_dims[i + 1], scale=layer_scale
                )
                for i in range(depth - 1)
            ]
        )

        self.up_layers = nn.ModuleList(
            [
                PromptSynthesisBlock(
                    channel_dims[-i],
                    channel_dims[-i - 1],
                    channel_dims[-i - 1],
                    scale=layer_scale,
                )
                for i in range(1, depth)
            ]
        )

        self.prompt_out = nn.Conv3d(
            channel_dims[0], 1, kernel_size=1, padding="same"
        )
        self.box_out = PromptBoxPredictor(channel_dims[0])

    def forward(self, x: Tensor, num_batches: int) -> tuple[Tensor, Tensor]:
        BD, C, H, W = x.shape
        x = x.view(num_batches, -1, C, H, W).transpose(1, 2)  # (B, C, D, H, W)
        x = self.init_conv(x)  # (B, C', D, H, W)

        outputs = []

        for block in self.down_layers:
            outputs.append(x)
            x = block(x)
        for layer, skip_x in zip(
            self.up_layers, reversed(outputs), strict=False
        ):
            x = layer(x, skip_x)
        prompt_outs = self.prompt_out(x)  # (B, 1, D, H, W)
        prompt_outs = prompt_outs.view(BD, 1, H, W)  # (B*D, 1, H, W)
        prompt_outs = F.interpolate(
            prompt_outs,
            scale_factor=self.scale_factor,
            mode="bilinear",
            align_corners=True,
        )  # Upscale to original size
        prompt_outs = torch.sigmoid(prompt_outs)  # Normalize to [0, 1]
        box_outs = self.box_out(x)  # (B*D, 4)
        return box_outs, prompt_outs


class LoRALinear(nn.Module):
    """A linear layer with LoRA (Low-Rank Adaptation) applied."""

    def __init__(
        self, proj: nn.Module, input_dim: int, output_dim: int, r: int, a: int
    ):
        super().__init__()
        self.proj = proj
        self.r = r
        self.a = a
        self.scaling = a / r
        self.w_a = nn.Linear(input_dim, r, bias=False)
        self.w_b = nn.Linear(r, output_dim, bias=False)

        self.initialize_parameters()

    def initialize_parameters(self):
        """Initializes LoRA parameters."""
        nn.init.kaiming_uniform_(self.w_a.weight, a=np.sqrt(5))
        nn.init.zeros_(self.w_b.weight)

    def forward(self, x: Tensor) -> Tensor:
        return self.proj(x) + self.w_b(self.w_a(x)) * self.scaling


class LoRAMaskDecoderFactory:
    """Module to apply LoRA to the transformer blocks in the SAM MaskDecoder."""

    def __init__(self, lora_r: int = 32, lora_alpha: int = 64):
        self.r = lora_r
        self.a = lora_alpha

    def _apply_lora(self, attn_block: Attention) -> None:
        """Applies LoRA to the attention block."""
        in_features = attn_block.embedding_dim
        out_features = attn_block.internal_dim

        # Original projection layer
        q_proj = attn_block.q_proj
        v_proj = attn_block.v_proj

        # Initialize LoRA layers
        attn_block.q_proj = LoRALinear(q_proj, in_features, out_features, self.r, self.a)  # type: ignore
        attn_block.v_proj = LoRALinear(v_proj, in_features, out_features, self.r, self.a)  # type: ignore

    def apply(self, mask_decoder: nn.Module) -> nn.Module:
        """Applies LoRA to all transformer blocks in the MaskDecoder."""

        for p in mask_decoder.parameters():
            p.requires_grad = False

        transformer = mask_decoder.transformer

        for _, blk in enumerate(transformer.layers):  # type: ignore
            self._apply_lora(blk.self_attn)
            self._apply_lora(blk.cross_attn_token_to_image)
            self._apply_lora(blk.cross_attn_image_to_token)

        self._apply_lora(transformer.final_attn_token_to_image)  # type: ignore

        return mask_decoder



from pathlib import Path
from typing import Annotated

from typer import Argument, Option

from .cli import cli


@cli.command(name="features", no_args_is_help=True)
def features(
    tomograms: Annotated[
        str,
        Argument(
            help="Path to the folder or .txt file containing the tomograms to process.",
        ),
    ],
    result_folder: Annotated[
        str,
        Argument(
            help="Path to the folder where the DINO features will be saved.",
        ),
    ],
    batch_size: Annotated[
        int,
        Option(
            min=1,
            help="Batch size for DINO feature extraction.",
            rich_help_panel="Customization and Utils",
        ),
    ] = 64,
    window_size: Annotated[
        int | None,
        Option(
            min=1,
            help="Window size for DINO feature extraction.",
            show_default="720",
            rich_help_panel="Customization and Utils",
        ),
    ] = None,
    visualize: Annotated[
        bool,
        Option(
            "--visualize",
            "-v",
            help="Save PCA visualization of DINO features? This will increase the runtime.",
            rich_help_panel="Customization and Utils",
        ),
    ] = False,
):
    """Compute high-level features using DINOv2 for a set of tomograms.

    Example
    -------
    cryovit features <path-to-tomograms> <path-to-result-folder>
    """
    from cryovit._logging_config import setup_logging
    from cryovit.config import DEFAULT_WINDOW_SIZE, DINO_PATCH_SIZE
    from cryovit.run.dino_features import run_dino
    from cryovit.utils import load_files_from_path

    setup_logging("INFO")

    ## Convert Arguments
    tomograms_path = Path(tomograms)
    result_path = Path(result_folder)
    window_size = (
        (DINO_PATCH_SIZE * window_size // 16)
        if window_size is not None
        else DEFAULT_WINDOW_SIZE
    )

    ## Sanity Checking
    assert tomograms_path.exists(), "Tomograms path does not exist."
    result_path.mkdir(parents=True, exist_ok=True)

    ## Run DINO feature extraction
    tomogram_files = load_files_from_path(tomograms_path)
    run_dino(
        tomogram_files,
        result_path,
        batch_size=batch_size,
        window_size=window_size,
        visualize=visualize,
    )



"""Module defining data loading functionality for running CryoViT on user datasets."""

import logging
from collections.abc import Callable
from pathlib import Path

from pytorch_lightning import LightningDataModule
from torch.utils.data import DataLoader

from cryovit.datamodules.utils import collate_fn
from cryovit.types import FileData


class FileDataModule(LightningDataModule):
    """Module defining common functions for creating data loaders for file-based datasets."""

    def __init__(
        self,
        data_paths: list[Path],
        dataset_fn: Callable,
        dataloader_fn: Callable,
        val_paths: list[Path] | None = None,
        data_labels: list[Path] | None = None,
        val_labels: list[Path] | None = None,
        labels: list[str] | None = None,
        **kwargs,
    ) -> None:
        """Initializes the BaseDataModule with dataset parameters, a dataloader function, and a path to the split file.

        Args:
            data_paths (list[Path]): A list of paths to the data files for training/testing/prediction.
            dataset_fn (Callable): Function to create a Dataset from a list of FileData objects.
            dataloader_fn (Callable): Function to create a DataLoader from a dataset.
            val_paths (Optional[list[Path]]): A list of paths to the data files for validation.
            data_labels (Optional[list[Path]]): A list of paths to the label files for training/testing/prediction. Should only be missing for inference.
            val_labels (Optional[list[Path]]): A list of paths to the label files for validation.
            labels (Optional[list[str]]): A list of label keys to load from the label files. Should only be missing if no labels are provided.
        """

        super().__init__()
        self.data_files = self._combine_files_and_labels(
            data_paths, data_labels, labels
        )
        self.val_files = (
            self._combine_files_and_labels(val_paths, val_labels, labels)
            if val_paths is not None
            else []
        )
        self.dataset_fn = dataset_fn
        self.dataloader_fn = dataloader_fn

    def _combine_files_and_labels(
        self,
        files: list[Path],
        labels: list[Path] | None,
        label_keys: list[str] | None,
    ) -> list[FileData]:
        """Combines data files and label files into a list of FileData objects. Replaces missing labels with None."""

        file_labels = [None] * len(files) if labels is None else labels
        if len(files) != len(file_labels):
            raise ValueError(
                "Number of data files must match number of label files."
            )
        combined = []
        for fp, lp in zip(files, file_labels, strict=True):
            if not fp.exists() or (lp is not None and not lp.exists()):
                logging.warning(
                    "File %s or label %s does not exist, skipping.", fp, lp
                )
                continue
            combined.append(
                FileData(
                    tomo_path=fp,
                    label_path=lp,
                    sample=fp.parent.name,
                    labels=label_keys,
                )
            )
        return combined

    def train_dataloader(self) -> DataLoader:
        """Creates DataLoader for training data.

        Returns:
            DataLoader: A DataLoader instance for training data.
        """

        if len(self.data_files) == 0:
            raise ValueError("No training data provided.")
        dataset = self.dataset_fn(self.data_files, train=True)
        return self.dataloader_fn(dataset, shuffle=True, collate_fn=collate_fn)

    def val_dataloader(self) -> DataLoader:
        """Creates DataLoader for validation data.

        Returns:
            DataLoader: A DataLoader instance for validation data.
        """

        if len(self.val_files) == 0:
            logging.warning(
                "No validation data provided, using training data."
            )
            val_files = self.data_files
        else:
            val_files = self.val_files
        dataset = self.dataset_fn(val_files, train=False)
        return self.dataloader_fn(
            dataset, shuffle=False, collate_fn=collate_fn
        )

    def test_dataloader(self) -> DataLoader:
        """Creates DataLoader for testing data.

        Returns:
            DataLoader: A DataLoader instance for testing data.
        """

        if len(self.data_files) == 0:
            raise ValueError("No testing data provided.")
        dataset = self.dataset_fn(self.data_files, train=False)
        return self.dataloader_fn(
            dataset, shuffle=False, collate_fn=collate_fn
        )

    def predict_dataloader(self) -> DataLoader:
        """Creates DataLoader for prediction data.

        Returns:
            DataLoader: A DataLoader instance for prediction data.
        """

        if len(self.data_files) == 0:
            raise ValueError("No prediction data provided.")
        dataset = self.dataset_fn(self.data_files, train=False)
        return self.dataloader_fn(
            dataset, shuffle=False, collate_fn=collate_fn
        )



"""Setup formatted logging with RichHandler."""

import logging

from rich.logging import RichHandler


def setup_logging(level: str = "INFO") -> None:
    logger = logging.getLogger()
    logger.setLevel(level)

    if not logger.hasHandlers():
        rich_handler = RichHandler(rich_tracebacks=True)
        rich_handler.setLevel(level)
        formatter = logging.Formatter("%(message)s", datefmt="[%X]")
        rich_handler.setFormatter(formatter)
        logger.addHandler(rich_handler)



"""Make plots comparing fractional LOO sample performance."""

import functools
from pathlib import Path

import pandas as pd

from cryovit.visualization.utils import (
    compute_stats,
    merge_experiments,
    significance_test,
)


def _plot_df(
    df: pd.DataFrame,
    pvalues: pd.Series,
    key: str,
    title: str,
    file_name: str,
    plot_points: bool = True,
):
    import matplotlib
    import matplotlib.pyplot as plt
    import seaborn as sns
    from statannotations.Annotator import Annotator

    matplotlib.use("Agg")
    colors = sns.color_palette("deep")[:3]
    sns.set_theme(style="darkgrid", font="Open Sans")

    hue_palette = {
        "3D U-Net": colors[0],
        "CryoViT": colors[1],
    }

    figsize = (20, 6) if "Mito" in title else (10, 6)
    fig = plt.figure(figsize=figsize)
    ax = plt.gca()
    df["split_id"] = df["split_id"].astype(int)
    label_counts = df["split_id"].value_counts()
    num_models = df[key].nunique()

    params = {
        "x": "split_id",
        "y": "dice_metric",
        "hue": key,
        "data": df,
    }

    if plot_points:
        sns.boxplot(
            showfliers=False,
            palette=hue_palette,
            linewidth=1,
            medianprops={"linewidth": 2, "color": "firebrick"},
            ax=ax,
            **params,
        )
        sns.stripplot(
            dodge=True,
            marker=".",
            alpha=0.5,
            palette="dark:black",
            ax=ax,
            **params,
        )

        k1, k2 = df[key].unique()
        pairs = [[(s, k1), (s, k2)] for s in pvalues.index]

        annotator = Annotator(ax, pairs, **params)
        annotator.configure(color="blue", line_width=1, verbose=False)
        annotator.set_pvalues_and_annotate(pvalues.values)
    else:
        sns.lineplot(
            marker="o",
            markersize=8,
            linewidth=2,
            palette=hue_palette,
            ax=ax,
            **params,
        )
        ax.set_xticks(df["split_id"].unique())

    current_labels = ax.get_xticklabels()
    new_labels = [f"{label.get_text()}0%" for label in current_labels]
    ax.set_xticks(ax.get_xticks())
    ax.set_xticklabels(new_labels, ha="center")

    ax.set_ylim(-0.05, 1.15)
    ax.set_xlabel("")
    ax.set_ylabel("")

    fig.suptitle(title)
    fig.supxlabel(
        f"Fraction of Training Data\n(n={label_counts[10] // num_models})"
    )
    fig.supylabel("Dice Score")

    handles, labels = ax.get_legend_handles_labels()
    plt.legend(handles[:2], labels[:2], loc="lower center", shadow=True)

    plt.tight_layout()
    plt.savefig(f"{file_name}{'_line' if not plot_points else ''}.svg")
    plt.savefig(
        f"{file_name}{'_line' if not plot_points else ''}.png", dpi=300
    )


def process_fractional_experiment(
    exp_type: str,
    label: str,
    exp_names: dict[str, list[str]],
    exp_dir: Path,
    result_dir: Path,
    plot_points: bool = True,
):
    """Plot fractional experiment results with box and strip plots including annotations for statistical tests.

    Args:
        exp_type (str): Type of experiment, e.g., "fractional", "sparse"
        label (str): The label being analyzed, e.g., "mito", "cristae"
        exp_names (dict[str, list[str]]): Dictionary mapping experiment names to model used
        exp_dir (Path): Directory containing the experiment results
        result_dir (Path): Directory to save the results
    """

    df = merge_experiments(exp_dir, exp_names, keys=["model"])
    test_fn = functools.partial(
        significance_test,
        model_A=("CryoViT"),
        model_B=("3D U-Net"),
        key="model",
        test_fn="ttest_rel",
    )
    full_labels = {
        "mito": "Mitochondria",
        "cristae": "Cristae",
        "microtubule": "Microtubules",
        "granule": "Granules",
        "bacteria": "Bacteria",
    }
    p_values = compute_stats(
        df,
        group_keys=["split_id", "model"],
        file_name=str(result_dir / f"{label}_{exp_type}_stats.csv"),
        test_fn=test_fn,
    )
    _plot_df(
        df,
        p_values,
        "model",
        f"Fractional Model Comparison for {full_labels[label]}",
        str(result_dir / f"{label}_{exp_type}_comparison"),
        plot_points=plot_points,
    )



"""Make plots comparing single sample performance."""

import functools
from pathlib import Path

import pandas as pd

from cryovit.types import Sample
from cryovit.visualization.utils import (
    compute_stats,
    merge_experiments,
    significance_test,
)


def _plot_df(
    df: pd.DataFrame,
    pvalues: pd.Series,
    key: str,
    title: str,
    file_name: str,
    is_sample: bool,
):
    # import here to avoid unnecessary dependencies if function not used
    import matplotlib
    import matplotlib.pyplot as plt
    import seaborn as sns
    from statannotations.Annotator import Annotator

    matplotlib.use("Agg")
    colors = sns.color_palette("deep")[:3]
    sns.set_theme(style="darkgrid", font="Open Sans")

    hue_palette = {
        "3D U-Net": colors[0],
        "CryoViT": colors[1],
        "CryoViT with Sparse Labels": colors[1],
        "CryoViT with Dense Labels": colors[2],
    }

    fig = plt.figure(figsize=(20, 6))
    ax = plt.gca()
    if not is_sample:
        df["split_id"] = df["split_id"].astype(int)
    label_key = "sample" if is_sample else "split_id"
    label_counts = df[label_key].value_counts()
    num_models = df[key].nunique()
    sorted_labels = label_counts.sort_values(ascending=True).index.tolist()

    params = {
        "x": label_key,
        "y": "dice_metric",
        "hue": key,
        "data": df,
        "order": sorted_labels,
    }

    sns.boxplot(
        showfliers=False,
        palette=hue_palette,
        linewidth=1,
        medianprops={"linewidth": 2, "color": "firebrick"},
        ax=ax,
        **params,
    )
    sns.stripplot(
        dodge=True,
        marker=".",
        alpha=0.5,
        palette="dark:black",
        ax=ax,
        **params,
    )

    k1, k2 = df[key].unique()
    pairs = [[(s, k1), (s, k2)] for s in pvalues.index]

    annotator = Annotator(ax, pairs, **params)
    annotator.configure(color="blue", line_width=1, verbose=False)
    annotator.set_pvalues_and_annotate(pvalues.values)

    current_labels = ax.get_xticklabels()
    if is_sample:
        new_labels = [
            f"{Sample(label.get_text()).value}\n(n={label_counts[label.get_text()] // num_models})"
            for label in current_labels
        ]
    else:
        new_labels = [f"{label.get_text()}0%" for label in current_labels]
    ax.set_xticks(ax.get_xticks())
    ax.set_xticklabels(new_labels, ha="center")

    ax.set_ylim(-0.05, 1.15)
    ax.set_xlabel("")
    ax.set_ylabel("")

    fig.suptitle(title)
    fig.supxlabel(
        "Sample Name (Count)"
        if is_sample
        else f"Fraction of Training Data\n(n={label_counts[10] // num_models})"
    )
    fig.supylabel("Dice Score")

    handles, labels = ax.get_legend_handles_labels()
    plt.legend(handles[:2], labels[:2], loc="lower center", shadow=True)

    plt.tight_layout(rect=(0.01, 0.01, 1.0, 1.0))
    plt.savefig(f"{file_name}.svg")
    plt.savefig(f"{file_name}.png", dpi=300)


def process_sparse_experiment(
    exp_type: str,
    exp_group: str,
    exp_names: dict[str, list[str]],
    exp_dir: Path,
    result_dir: Path,
):
    """Plot sparse sample experiment results with box and strip plots including annotations for statistical tests.

    Args:
        exp_type (str): Type of experiment, e.g. "sparse"
        exp_group (str): Group of experiments, e.g. "single", "fractional"
        exp_names (dict[str, list[str]]): Dictionary mapping experiment names to models
        exp_dir (Path): Directory containing experiment results
        result_dir (Path): Directory to save results
    """

    result_dir.mkdir(parents=True, exist_ok=True)
    df = merge_experiments(exp_dir, exp_names, keys=["label_type"])
    if exp_group == "single":
        df = df[df["split_id"] == 10]  # Use data from 100% of training data
    test_type = "wilcoxon" if exp_group == "single" else "ttest_rel"
    test_fn = functools.partial(
        significance_test,
        model_A=("CryoViT with Sparse Labels"),
        model_B=("CryoViT with Dense Labels"),
        key="label_type",
        test_fn=test_type,
    )
    group_key = "sample" if exp_group == "single" else "split_id"
    p_values = compute_stats(
        df,
        group_keys=[group_key, "label_type"],
        file_name=str(result_dir / f"{exp_group}_{exp_type}_stats.csv"),
        test_fn=test_fn,
    )
    title = f"CryoViT: Sparse vs. Dense Label Comparison on {'Individual' if exp_group == 'single' else 'Fractional'} Samples"
    _plot_df(
        df,
        p_values,
        "label_type",
        title,
        str(result_dir / f"{exp_group}_{exp_type}_comparison"),
        is_sample=(exp_group == "single"),
    )



"""UNet3D model architecture for 3D tomogram segmentation."""

import math

import torch
from torch import Tensor, nn

from cryovit.models.base_model import BaseModel
from cryovit.types import BatchedTomogramData


class UNet3D(BaseModel):
    """UNet3D model implementation."""

    def __init__(self, **kwargs) -> None:
        """Initializes the UNet3D model with specific analysis and synthesis blocks."""

        super().__init__(**kwargs)

        self.bottom_layer = nn.Sequential(
            nn.Conv3d(256, 384, 3, padding="same"),
            nn.InstanceNorm3d(384, eps=1e-3, affine=True),
            nn.GELU(),
            nn.Conv3d(384, 256, 3, padding="same"),
            nn.InstanceNorm3d(256, eps=1e-3, affine=True),
            nn.GELU(),
        )

        self.analysis_layers = nn.ModuleList(
            [
                AnalysisBlock(1, 16),
                AnalysisBlock(16, 64),
                AnalysisBlock(64, 256),
            ]
        )

        self.synthesis_layers = nn.ModuleList(
            [
                SynthesisBlock(256, 256, 64),
                SynthesisBlock(64, 64, 16),
                SynthesisBlock(16, 16, 16),
            ]
        )

        self.output_layer = nn.Conv3d(16, 1, 1, padding="same")
        self.PAD = max(16, 2 ** len(self.analysis_layers))

    def forward_volume(self, x: Tensor) -> Tensor:
        """Memory optimized forward pass for the UNet3D model."""

        analysis_outputs = []

        for block in self.analysis_layers:
            x, prev_x = block(x)
            analysis_outputs.append(prev_x)

        for layer in self.bottom_layer:
            x = layer(x)

        for block in self.synthesis_layers:
            for layer in block.upconv:  # type: ignore
                x = layer(x)

            x = torch.cat([x, analysis_outputs.pop()], 1)

            for layer in block.layers:  # type: ignore
                x = layer(x)

        x = self.output_layer(x)
        x = torch.clip(x, -5.0, 5.0)
        return x

    def forward(self, batch: BatchedTomogramData) -> Tensor:  # type: ignore
        x = batch.tomo_batch  # (B, D, C, H, W)
        x = x.permute(0, 2, 1, 3, 4)  # (B, C, D, H, W)

        # Add padding
        new_size = [
            self.PAD * math.ceil(dim / self.PAD) for dim in x.size()[-3:]
        ]
        D, H, W = x.size()[-3:]
        if new_size != [D, H, W]:
            x = self._add_padding(x, new_size)

        x = self.forward_volume(x)  # (B, 1, D, H, W)

        # Remove padding
        if new_size != [D, H, W]:
            x = x[..., :D, :H, :W]

        x = x.squeeze(1)  # (B, D, H, W)
        return torch.sigmoid(x)

    @torch.inference_mode()
    def _add_padding(self, x: Tensor, new_size: list[int]) -> Tensor:
        """Adds padding to the input to match the U-Net dimensions"""

        # pad to multiple of self.PAD
        D, H, W = x.size()[-3:]
        new_shape = list(x.shape[:-3]) + new_size
        x_new = torch.zeros(*new_shape, dtype=x.dtype, device=x.device)
        x_new[..., :D, :H, :W] = x

        return x_new


class AnalysisBlock(nn.Module):
    """Block that performs down-sampling and feature extraction in UNet3D's analysis path."""

    def __init__(self, in_channels: int, out_channels: int) -> None:
        """Initializes the AnalysisBlock with convolutional layers and pooling.

        Args:
            in_channels (int): Number of input channels.
            out_channels (int): Number of output channels.
        """

        super().__init__()
        self.pool = nn.Sequential(
            nn.Conv3d(out_channels, out_channels, 2, stride=2),
            nn.InstanceNorm3d(out_channels, eps=1e-3, affine=True),
            nn.GELU(),
        )

        self.layers = nn.Sequential(
            nn.Conv3d(in_channels, out_channels, 3, padding="same"),
            nn.InstanceNorm3d(out_channels, eps=1e-3, affine=True),
            nn.GELU(),
            nn.Conv3d(out_channels, out_channels, 3, padding="same"),
            nn.InstanceNorm3d(out_channels, eps=1e-3, affine=True),
            nn.GELU(),
        )

    def forward(self, x: Tensor) -> tuple[Tensor, Tensor]:
        """Forward pass for the analysis block."""

        x = self.layers(x)
        y = self.pool(x)
        return y, x


class SynthesisBlock(nn.Module):
    """Block that performs up-sampling and feature combination in UNet3D's synthesis path."""

    def __init__(
        self, in_channels: int, skip_channels: int, out_channels: int
    ) -> None:
        """Initializes the SynthesisBlock with up-sampling and convolutional layers.

        Args:
            in_channels (int): Number of input channels.
            skip_channels (int): Number of channels from the skip connection.
            out_channels (int): Number of output channels.
        """

        super().__init__()
        self.upconv = nn.Sequential(
            nn.ConvTranspose3d(in_channels, out_channels, 2, stride=2),
            nn.InstanceNorm3d(out_channels, eps=1e-3, affine=True),
            nn.GELU(),
        )

        self.layers = nn.Sequential(
            LinearProjection(out_channels + skip_channels, out_channels),
            nn.InstanceNorm3d(out_channels, eps=1e-3, affine=True),
            nn.GELU(),
            nn.Conv3d(out_channels, out_channels, 3, padding="same"),
            nn.InstanceNorm3d(out_channels, eps=1e-3, affine=True),
            nn.GELU(),
        )

    def forward(self, x: Tensor, skip_x: Tensor) -> Tensor:
        """Forward pass for the synthesis block."""

        x = self.upconv(x)
        x = torch.cat([x, skip_x], 1)  # channel concat
        x = self.layers(x)
        return x


class LinearProjection(nn.Module):
    """Linear projection of an input tensor."""

    def __init__(self, in_channels: int, out_channels: int) -> None:
        """Initializes the LinearProjection with a linear transformation.

        Args:
            in_channels (int): Number of input channels.
            out_channels (int): Number of output channels after projection.
        """

        super().__init__()
        self.proj = nn.Linear(in_channels, out_channels)

    def forward(self, x: Tensor) -> Tensor:
        """Forward pass for the linear projection layer."""

        x = torch.permute(x, (0, 2, 3, 4, 1))
        x = self.proj(x)
        x = torch.permute(x, (0, 4, 1, 2, 3))
        return x



"""Make plots comparing multi sample performance."""

import functools
from pathlib import Path

import pandas as pd

from cryovit.types import Sample
from cryovit.visualization.utils import (
    compute_stats,
    merge_experiments,
    significance_test,
)

group_names = {
    "hd": "Diseased",
    "healthy": "Healthy",
    "old": "Aged",
    "young": "Young",
    "neuron": "Neurons",
    "fibro_cancer": "Fibroblast/Cancer Cells",
}


def _plot_df(
    df: pd.DataFrame,
    pvalues: pd.Series,
    key: str,
    title: str,
    ax,
):
    import matplotlib
    import seaborn as sns
    from statannotations.Annotator import Annotator

    matplotlib.use("Agg")
    colors = sns.color_palette("deep")[:2]
    sns.set_theme(style="darkgrid", font="Open Sans")

    hue_palette = {
        "3D U-Net": colors[0],
        "CryoViT": colors[1],
    }

    sample_counts = df["sample"].value_counts()
    num_models = df[key].nunique()
    sorted_samples = sample_counts.sort_values(ascending=True).index.tolist()

    params = {
        "x": "sample",
        "y": "dice_metric",
        "hue": key,
        "data": df,
        "order": sorted_samples,
    }

    sns.boxplot(
        showfliers=False,
        palette=hue_palette,
        linewidth=1,
        medianprops={"linewidth": 2, "color": "firebrick"},
        ax=ax,
        **params,
    )
    sns.stripplot(
        dodge=True,
        marker=".",
        alpha=0.5,
        palette="dark:black",
        ax=ax,
        **params,
    )

    k1, k2 = df[key].unique()
    pairs = [[(s, k1), (s, k2)] for s in pvalues.index]

    annotator = Annotator(ax, pairs, **params)
    annotator.configure(color="blue", line_width=1, verbose=False)
    annotator.set_pvalues_and_annotate(pvalues.values)

    current_labels = ax.get_xticklabels()
    new_labels = [
        f"{Sample[label.get_text()].value}\n(n={sample_counts[label.get_text()] // num_models})"
        for label in current_labels
    ]

    ax.set_ylim(-0.05, 1.15)
    ax.set_title(title)
    ax.set_xlabel("")
    ax.set_ylabel("")
    ax.set_xticks(range(len(new_labels)))
    ax.set_xticklabels(new_labels, ha="center")

    handles, labels = ax.get_legend_handles_labels()
    ax.legend(handles[:2], labels[:2], loc="lower right", shadow=True)


def process_multi_experiment(
    exp_type: str,
    exp_group: tuple[str, str],
    exp_names: dict[str, list[str]],
    exp_dir: Path,
    result_dir: Path,
):
    """Plot domain-shift results with box and strip plots including annotations for statistical tests.

    Args:
        exp_type (str): Type of experiment, i.e., "multi"
        exp_group (tuple[str, str]): Tuple containing the names of the two experiment groups to compare
        exp_names (dict[str, list[str]]): Dictionary mapping experiment group names to a model and comparison direction, e.g., "forward" or "backward"
        exp_dir (Path): Directory containing the experiment results
        result_dir (Path): Directory to save the results
    """

    import matplotlib.pyplot as plt
    import seaborn as sns
    from matplotlib.gridspec import GridSpec

    sns.set_theme(style="darkgrid", font="Open Sans")

    result_dir.mkdir(parents=True, exist_ok=True)
    df = merge_experiments(exp_dir, exp_names, keys=["model", "type"])
    forward_df = df[df["type"] == "forward"]
    backward_df = df[df["type"] == "backward"]

    s1_count = forward_df["sample"].nunique()
    s2_count = backward_df["sample"].nunique()

    fig = plt.figure(figsize=(20, 6))
    gs = GridSpec(
        1, 2, width_ratios=[s1_count, s2_count]
    )  # Set width ratios based on unique sample counts

    ax1 = fig.add_subplot(gs[0])
    ax2 = fig.add_subplot(gs[1])

    # Plot forward comparison (s1 vs. s2)
    test_fn = functools.partial(
        significance_test,
        model_A="CryoViT",
        model_B="3D U-Net",
        key="model",
        test_fn="wilcoxon",
    )
    p_values = compute_stats(
        forward_df,
        group_keys=["sample", "model"],
        file_name=str(
            result_dir / f"{'_'.join(list(exp_group))}_{exp_type}_stats.csv"
        ),
        test_fn=test_fn,
    )
    title = f"{group_names[exp_group[0]]} to {group_names[exp_group[1]]} Shift"
    _plot_df(forward_df, p_values, "model", title, ax1)

    # Plot backward comparison (s2 vs. s1)
    test_fn = functools.partial(
        significance_test,
        model_A="CryoViT",
        model_B="3D U-Net",
        key="model",
        test_fn="wilcoxon",
    )
    p_values = compute_stats(
        backward_df,
        group_keys=["sample", "model"],
        file_name=str(
            result_dir
            / f"{'_'.join(list(reversed(exp_group)))}_{exp_type}_stats.csv"
        ),
        test_fn=test_fn,
    )
    title = f"{group_names[exp_group[1]]} to {group_names[exp_group[0]]} Shift"
    _plot_df(backward_df, p_values, "model", title, ax2)

    # Adjust layout and save the figure
    if (
        "Cells" in group_names[exp_group[0]]
        or "Cells" in group_names[exp_group[1]]
    ):
        domain = "Cell Type"
    else:
        domain = "Diseased/Healthy"
    fig.suptitle(f"Model Comparison Across {domain} Domain Shifts")
    fig.supxlabel("Sample Name (Count)")
    fig.supylabel("Dice Score")

    plt.tight_layout(rect=(0.01, 0.01, 1.0, 1.0))
    plt.savefig(result_dir / f"{exp_group[0]}_{exp_group[1]}_domain_shift.svg")
    plt.savefig(
        result_dir / f"{exp_group[0]}_{exp_group[1]}_domain_shift.png", dpi=300
    )



"""CryoVIT model architecture for 3D tomogram segmentation."""

import torch
from torch import Tensor, nn

from cryovit.models.base_model import BaseModel
from cryovit.types import BatchedTomogramData


class CryoVIT(BaseModel):
    """CryoVIT model implementation."""

    def __init__(self, **kwargs) -> None:
        """Initializes the CryoVIT model with specific convolutional and synthesis blocks."""

        super().__init__(**kwargs)

        self.layers = nn.Sequential(
            nn.Conv3d(
                1536, 1024, 1, padding="same"
            ),  # projection to a lower dimension
            nn.GELU(),
            SynthesisBlock(1024, 192, 128, d1=32, d2=24),
            SynthesisBlock(128, 64, 32, d1=16, d2=12),
            SynthesisBlock(32, 32, 32, d1=8, d2=4),
            SynthesisBlock(32, 16, 8, d1=2, d2=1),
        )

        # output layer for generating the final segmentation
        self.output_layer = nn.Sequential(
            nn.Conv3d(8, 8, 3, padding="same"),
            nn.GELU(),
            nn.Conv3d(8, 1, 3, padding="same"),
        )

    def forward_volume(self, x: Tensor) -> Tensor:
        x = self.layers(x)
        x = self.output_layer(x)
        x = torch.clip(x, -5.0, 5.0)
        return x

    def forward(self, batch: BatchedTomogramData) -> Tensor:  # type: ignore
        """Forward pass for the CryoVIT model."""

        x = batch.tomo_batch  # (B, D, C, H, W)
        x = x.permute(0, 2, 1, 3, 4)  # (B, C, D, H, W)
        x = self.forward_volume(x)
        x = x.squeeze(1)  # (B, D, H, W)
        return torch.sigmoid(x)


class SynthesisBlock(nn.Module):
    """Synthesis block for anisotropic upscaling with dilated convolutions."""

    def __init__(self, c1: int, c2: int, c3: int, d1: int, d2: int) -> None:
        """Initializes the Synthesis block for anisotropic upscaling.

        Args:
            c1 (int): Number of channels in the input volume.
            c2 (int): Number of channels in the intermediate tensor.
            c3 (int): Number of channels in the output upscaled volume.
            d1 (int): Depthwise dilation rate for the first 3D Conv layer.
            d2 (int): Depthwise dilation rate for the second 3D Conv layer.
        """

        super().__init__()

        self.layers = nn.Sequential(
            nn.GroupNorm(max(8, c1 // 8), c1, eps=1e-3),
            nn.Conv3d(c1, c2, 3, padding="same", dilation=(d1, 1, 1)),
            nn.GELU(),
            nn.Conv3d(c2, c2, 3, padding="same", dilation=(d2, 1, 1)),
            nn.GELU(),
            nn.ConvTranspose3d(
                c2, c3, (1, 2, 2), stride=(1, 2, 2)
            ),  # upscale by 2
            nn.GELU(),
        )

    def forward(self, x: Tensor) -> Tensor:
        """Forward pass for the synthesis block."""

        return self.layers(x)



"""Script to train segmentation models for CryoET data."""

import traceback
import warnings

import hydra
import torch

from cryovit._logging_config import setup_logging
from cryovit.config import BaseExperimentConfig, validate_experiment_config
from cryovit.run import train_model

setup_logging("DEBUG")
import logging  # noqa: E402

warnings.simplefilter("ignore")


@hydra.main(
    config_path="../configs",
    config_name="train_model",
    version_base="1.2",
)
def main(cfg: BaseExperimentConfig) -> None:
    """Main function to orchestrate the training of segmentation models.

    Validates the provided configuration, then initializes and runs the training process using the
    specified settings. Catches and logs any exceptions that occur during training.

    Args:
        cfg (TrainModelConfig): Configuration object loaded from train_model.yaml.
    """

    validate_experiment_config(cfg)
    result = 0
    try:
        train_model.run_trainer(cfg)
    except BaseException as err:  # noqa: BLE001
        logging.error("%s: %s", type(err).__name__, err)
        logging.error(traceback.format_exc())
        result = -1
    finally:
        import wandb

        # Free up GPU memory
        if torch.cuda.is_available():
            torch.cuda.empty_cache()
        # Ensure W&B finishes properly
        wandb.Settings(quiet=True)  # Disable W&B output
        if wandb.run is not None:
            wandb.finish(exit_code=result)


if __name__ == "__main__":
    main()



"""Make plots comparing single sample performance."""

import functools
from pathlib import Path

import pandas as pd

from cryovit.types import Sample
from cryovit.visualization.utils import (
    compute_stats,
    merge_experiments,
    significance_test,
)


def _plot_df(
    df: pd.DataFrame,
    pvalues: pd.Series,
    key: str,
    title: str,
    ax,
):
    # import here to avoid unnecessary dependencies if function not used
    import matplotlib
    import seaborn as sns
    from statannotations.Annotator import Annotator

    matplotlib.use("Agg")
    colors = sns.color_palette("deep")[:2]
    sns.set_theme(style="darkgrid", font="Open Sans")

    hue_palette = {
        "3D U-Net": colors[0],
        "CryoViT": colors[1],
    }

    sample_counts = df["sample"].value_counts()
    num_models = df[key].nunique()
    sorted_samples = sample_counts.sort_values(ascending=True).index.tolist()

    params = {
        "x": "sample",
        "y": "dice_metric",
        "hue": key,
        "data": df,
        "order": sorted_samples,
    }

    sns.boxplot(
        showfliers=False,
        palette=hue_palette,
        linewidth=1,
        medianprops={"linewidth": 2, "color": "firebrick"},
        ax=ax,
        **params,
    )
    sns.stripplot(
        dodge=True,
        marker=".",
        alpha=0.5,
        palette="dark:black",
        ax=ax,
        **params,
    )

    k1, k2 = df[key].unique()
    pairs = [[(s, k1), (s, k2)] for s in pvalues.index]

    annotator = Annotator(ax, pairs, **params)
    annotator.configure(color="blue", line_width=1, verbose=False)
    annotator.set_pvalues_and_annotate(pvalues.values)

    current_labels = ax.get_xticklabels()
    new_labels = [
        f"{Sample[label.get_text()].value}\n(n={sample_counts[label.get_text()] // num_models})"
        for label in current_labels
    ]

    ax.set_ylim(-0.05, 1.15)
    ax.set_title(title)
    ax.set_xlabel("")
    ax.set_ylabel("")
    ax.set_xticks(range(len(new_labels)))
    ax.set_xticklabels(new_labels, ha="center")

    handles, labels = ax.get_legend_handles_labels()
    if title == "HD":
        ax.legend(handles[:2], labels[:2], loc="lower center", shadow=True)
    else:
        ax.legend().remove()
    return handles, labels


def process_single_experiment(
    exp_type: str,
    exp_group: str,
    exp_names: dict[str, dict[str, list[str]]],
    exp_dir: Path,
    result_dir: Path,
):
    """Plot single sample experiment results with box and strip plots including annotations for statistical tests.

    Args:
        exp_type (str): Type of experiment, e.g. "sparse"
        exp_group (str): Group of experiments, e.g. "hd", "ad"
        exp_names (dict[str, dict[str, list[str]]]): Dictionary mapping experiment groups to experiment names and models
        exp_dir (Path): Directory containing experiment results
        result_dir (Path): Directory to save results
    """
    import matplotlib.pyplot as plt
    import seaborn as sns

    sns.set_theme(style="darkgrid", font="Open Sans")

    result_dir.mkdir(parents=True, exist_ok=True)
    dfs = {
        group: merge_experiments(exp_dir, exp_names[group], keys=["model", "group"])  # type: ignore
        for group in exp_names
    }
    exp_counts = [df["sample"].nunique() for df in dfs.values()]
    fig, axes = plt.subplots(
        1,
        len(dfs),
        figsize=(20, 6),
        sharey="row",
        gridspec_kw={"width_ratios": exp_counts},
    )

    for ax, (group, df) in zip(axes, dfs.items(), strict=True):
        test_fn = functools.partial(
            significance_test,
            model_A=("CryoViT"),
            model_B=("3D U-Net"),
            key="model",
            test_fn="wilcoxon",
        )
        p_values = compute_stats(
            df,
            group_keys=["sample", "model"],
            file_name=str(
                result_dir / f"{group.lower()}_{exp_type}_stats.csv"
            ),
            test_fn=test_fn,
        )
        title = f"{group}"
        _plot_df(
            df,
            p_values,
            "model",
            title,
            ax,
        )

    # Adjust layout and save the figure
    fig.suptitle("Model Comparison on Individual Samples for Mitochondria")
    fig.supxlabel("Sample Name (Count)")
    fig.supylabel("Dice Score")

    plt.tight_layout(rect=(0.01, 0.01, 1.0, 1.0))
    plt.savefig(result_dir / f"{exp_group.lower()}_{exp_type}.svg")
    plt.savefig(result_dir / f"{exp_group.lower()}_{exp_type}.png", dpi=300)



"""Implementations of PyTorch datasets for loading Cryo-EM tomograms."""

from cryovit.datasets.file_dataset import FileDataset
from cryovit.datasets.tomo_dataset import TomoDataset
from cryovit.datasets.vit_dataset import VITDataset

__all__ = [
    "VITDataset",
    "TomoDataset",
    "FileDataset",
]



"""Script for evaluating CryoVIT models based on configuration files."""

import logging
from collections.abc import Iterable
from pathlib import Path

import torch
from hydra import compose, initialize
from hydra.utils import instantiate
from pytorch_lightning import Trainer, seed_everything

from cryovit.config import BaseExperimentConfig
from cryovit.models import create_sam_model_from_weights
from cryovit.utils import load_model

torch.set_float32_matmul_precision("high")

## For Scripts


def run_evaluation(
    test_data: list[Path],
    test_labels: list[Path],
    labels: list[str],
    model_path: Path,
    result_dir: Path,
    visualize: bool = True,
) -> Path:
    """Run evaluation on the specified test data and labels, saving result metrics as a .csv file.

    Args:
        test_data (list[Path]): List of paths to the test tomograms.
        test_labels (list[Path]): List of paths to the test labels.
        labels (list[str]): List of label names to evaluate.
        model_path (Path): Path to the trained model file.
        result_dir (Path): Directory where the evaluation results will be saved.
        visualize (bool, optional): Whether to visualize the evaluation results. Defaults to True.

    Returns:
        Path: Path to the evaluation results file.
    """

    ## Get model information
    model, model_type, model_name, label_key = load_model(model_path)
    ## Setup hydra config
    with initialize(
        version_base="1.2",
        config_path="../configs",
        job_name="cryovit_eval",
    ):
        cfg = compose(
            config_name="eval_model",
            overrides=[
                f"name={model_name}",
                f"label_key={label_key}",
                f"model={model_type.value}",
                "additional_keys=[data]",
                "datamodule=file",
            ],
        )
    cfg.paths.model_dir = Path(__file__).parent.parent / "foundation_models"
    cfg.paths.results_dir = result_dir

    # Check input key
    if cfg.model.input_key != "dino_features":
        cfg.model.input_key = None  # find available data instead

    ## Setup dataset
    dataset_fn = instantiate(cfg.datamodule.dataset)
    dataloader_fn = instantiate(cfg.datamodule.dataloader)
    datamodule = instantiate(cfg.datamodule, _convert_="all")(
        data_paths=test_data,
        data_labels=test_labels,
        labels=labels,
        val_path=None,
        val_labels=None,
        dataloader_fn=dataloader_fn,
        dataset_fn=dataset_fn,
    )
    logging.info("Setup dataset.")

    ## Setup training
    callbacks = [instantiate(cb_cfg) for cb_cfg in cfg.callbacks.values()]
    # Remove pred_writer if visualize is False
    loggers = [
        instantiate(lg_cfg)
        for lg_name, lg_cfg in cfg.logger.items()
        if (visualize or lg_name != "test_pred_writer")
    ]
    trainer = instantiate(cfg.trainer, callbacks=callbacks, logger=loggers)

    logging.info("Starting testing.")
    trainer.test(model, datamodule=datamodule)

    # Load and return metrics path
    metrics_path = cfg.paths.results_dir / "results" / f"{model_name}.csv"
    return metrics_path


## For Experiments


def setup_exp_dir(cfg: BaseExperimentConfig) -> BaseExperimentConfig:
    """Setup the experiment directory structure."""

    # Convert paths to Paths
    cfg.paths.model_dir = Path(cfg.paths.model_dir)
    cfg.paths.data_dir = Path(cfg.paths.data_dir)
    cfg.paths.exp_dir = Path(cfg.paths.exp_dir)
    cfg.paths.results_dir = Path(cfg.paths.results_dir)

    if not isinstance(cfg.datamodule.sample, str) and isinstance(
        cfg.datamodule.sample, Iterable
    ):
        sample = "_".join(sorted(cfg.datamodule.sample))
    else:
        sample = cfg.datamodule.sample
    if not isinstance(cfg.datamodule.test_sample, str) and isinstance(
        cfg.datamodule.test_sample, Iterable
    ):
        test_sample = "_".join(sorted(cfg.datamodule.test_sample))
    else:
        test_sample = cfg.datamodule.test_sample

    new_exp_dir = cfg.paths.exp_dir / cfg.name / sample
    if cfg.datamodule.split_id is not None:
        new_exp_dir = new_exp_dir / f"split_{cfg.datamodule.split_id}"
    # Append test sample only for fractional sample experiments (test sample is separated for validation)
    if "Fractional" in cfg.datamodule._target_ and test_sample is not None:
        new_exp_dir = new_exp_dir / f"test_{test_sample}"

    cfg.paths.results_dir.mkdir(parents=True, exist_ok=True)
    assert (
        new_exp_dir.exists()
    ), f"Experiment directory {new_exp_dir} does not exist. Run training first."
    cfg.paths.exp_dir = new_exp_dir

    if cfg.ckpt_path is None:
        cfg.ckpt_path = new_exp_dir / "weights.pt"
    return cfg


def run_trainer(cfg: BaseExperimentConfig) -> None:
    """Sets up and executes the model evaluation using the specified configuration.

    Args:
        cfg (EvalModelConfig): Configuration object containing all settings for the evaluation process.
    """

    seed_everything(cfg.random_seed, workers=True)

    # Setup experiment directories
    cfg = setup_exp_dir(cfg)
    assert (
        cfg.ckpt_path is not None and cfg.ckpt_path.exists()
    ), f"{cfg.paths.exp_dir} does not contain a checkpoint."

    # Setup dataset
    dataset_fn = instantiate(cfg.datamodule.dataset)
    dataloader_fn = instantiate(cfg.datamodule.dataloader)
    split_file = cfg.paths.data_dir / cfg.paths.csv_name / cfg.paths.split_name
    datamodule = instantiate(cfg.datamodule, _convert_="all")(
        split_file=split_file,
        dataloader_fn=dataloader_fn,
        dataset_fn=dataset_fn,
    )
    logging.info("Setup dataset.")

    # Setup evaluation
    callbacks = [instantiate(cb_cfg) for cb_cfg in cfg.callbacks.values()]
    loggers = [instantiate(lg_cfg) for lg_cfg in cfg.logger.values()]
    trainer: Trainer = instantiate(
        cfg.trainer, callbacks=callbacks, logger=loggers
    )
    logging.info("Setup trainer.")
    if cfg.model._target_ == "cryovit.models.sam2.SAM2":
        # Load SAM2 pre-trained models
        model = create_sam_model_from_weights(
            cfg.model, cfg.paths.model_dir / cfg.paths.sam_name
        )
    else:
        model = instantiate(cfg.model)

    # Load model weights
    if cfg.ckpt_path.suffix == ".pt":
        model.load_state_dict(torch.load(cfg.ckpt_path))
    elif cfg.ckpt_path.suffix == ".ckpt":
        model = model.load_from_checkpoint(cfg.ckpt_path)
    else:
        raise ValueError(
            f"Unsupported checkpoint format: {cfg.ckpt_path.suffix}. Use .pt or .ckpt files."
        )

    logging.info("Setup model.")

    logging.info("Starting testing.")
    trainer.test(model, datamodule=datamodule)



"""Script for evaluating CryoVIT models based on configuration files."""

import logging
from pathlib import Path

import torch
from hydra import compose, initialize
from hydra.utils import instantiate

from cryovit.models.callbacks import PredictionWriter
from cryovit.utils import load_model

torch.set_float32_matmul_precision("high")

## For Scripts


def run_inference(
    data_files: list[Path],
    model_path: Path,
    result_dir: Path,
    threshold: float = 0.5,
) -> list[Path]:
    """Run inference on the specified data files and saves the results.

    Args:
        data_files (list[Path]): List of paths to the input data files.
        model_path (Path): Path to the trained model file.
        result_dir (Path): Directory where the inference results will be saved.
        threshold (float, optional): Threshold for binary classification. Defaults to 0.5.

    Returns:
        list[Path]: List of paths to the saved result files.
    """

    # Get model information
    model, model_type, model_name, label_key = load_model(model_path)
    assert model is not None, "Loaded model is None."
    ## Setup hydra config
    with initialize(
        version_base="1.2",
        config_path="../configs",
        job_name="infer_model",
    ):
        cfg = compose(
            config_name="infer_model",
            overrides=[
                f"name={model_name}",
                f"label_key={label_key}",
                f"model={model_type.value}",
                "datamodule=file",
            ],
        )
    cfg.paths.model_dir = Path(__file__).parent.parent / "foundation_models"
    cfg.paths.results_dir = result_dir

    # Check input key
    if cfg.model.input_key != "dino_features":
        cfg.model.input_key = None  # find available data instead

    ## Setup dataset
    dataset_fn = instantiate(cfg.datamodule.dataset)
    dataloader_fn = instantiate(cfg.datamodule.dataloader)
    datamodule = instantiate(cfg.datamodule, _convert_="all")(
        data_paths=data_files,
        val_paths=None,
        dataloader_fn=dataloader_fn,
        dataset_fn=dataset_fn,
    )
    logging.info("Setup dataset.")

    ## Setup training
    pred_writer = PredictionWriter(
        results_dir=result_dir, label_key=label_key, threshold=threshold
    )
    callbacks = [instantiate(cb_cfg) for cb_cfg in cfg.callbacks.values()]
    callbacks.append(pred_writer)
    loggers = []
    trainer = instantiate(cfg.trainer, callbacks=callbacks, logger=loggers)

    logging.info("Starting prediction.")
    trainer.predict(model, datamodule=datamodule)

    result_paths = pred_writer.result_paths
    return result_paths



"""Command-line interface for CryoViT."""

## Necessary imports to register CLI commands in pyproject.toml. DO NOT REMOVE!
from .cli import cli  # noqa: F401
from .dino_cli import features  # noqa: F401
from .eval_cli import evaluate  # noqa: F401
from .infer_cli import infer  # noqa: F401
from .train_cli import train  # noqa: F401



"""Dataset class for loading tomograms for DINOv2 models."""

import logging
from pathlib import Path

import h5py
import numpy as np
import torch
import torch.nn.functional as F
from numpy.typing import NDArray
from torch.utils.data import Dataset
from torchvision.transforms import Normalize

from cryovit.config import DINO_PATCH_SIZE

IMAGENET_DEFAULT_MEAN = (0.485, 0.456, 0.406)
IMAGENET_DEFAULT_STD = (0.229, 0.224, 0.225)


class VITDataset(Dataset):
    """Dataset class for Vision Transformer models, loading and processing tomograms."""

    def __init__(self, data_root: Path, records: list[str]) -> None:
        """Initializes a dataset object to load tomograms, applying normalization and resizing for DINOv2 models.

        Args:
            root (Path): Root directory where tomogram files are stored.
            records (list[str]): A list of strings representing paths to tomogram files in the root directory.
        """

        self.root = (
            data_root if isinstance(data_root, Path) else Path(data_root)
        )
        self.records = records
        self.transform = Normalize(IMAGENET_DEFAULT_MEAN, IMAGENET_DEFAULT_STD)
        self._printed_resize_warning = False

    def __len__(self) -> int:
        """Returns the number of tomograms in the dataset."""

        return len(self.records)

    def __getitem__(self, idx: int) -> torch.Tensor:
        """Retrieves a preprocessed tomogram tensor from the dataset by index.

        Args:
            idx (int): Index of the tomogram to retrieve.

        Returns:
            torch.Tensor: A tensor representing the normalized and resized tomogram.

        Raises:
            IndexError: If the index is out of the dataset's bounds.
        """

        if idx >= len(self):
            raise IndexError

        record = self.records[idx]
        data = self._load_tomogram(record)
        return self._transform(data)

    def _load_tomogram(self, record: str) -> NDArray[np.float32]:
        """Loads a tomogram from disk, assuming it is stored as an .hdf file in a `data` key.

        Args:
            record (str): The file path to the tomogram relative to the root directory.

        Returns:
            NDArray[np.uint8]: The loaded tomogram as a numpy array.
        """

        tomo_path = self.root / record

        with h5py.File(tomo_path) as fh:
            data: np.ndarray = fh["data"][()]  # type: ignore

        if data.dtype == np.uint8:
            data = data.astype(np.float32) / 255.0
        return data

    def _transform(self, data: NDArray[np.float32]) -> torch.Tensor:
        """Applies normalization and resizing transformations to the tomogram.

        Args:
            data (NDArray[np.float32]): The loaded tomogram data as a numpy array.

        Returns:
            torch.Tensor: The transformed data as a PyTorch tensor.
        """

        scale = (DINO_PATCH_SIZE / 16, DINO_PATCH_SIZE / 16)
        _, h, w = data.shape
        # Resize height and width to be multiples of 16
        H = int(np.ceil(h / 16) * 16)
        W = int(np.ceil(w / 16) * 16)
        if h != H or w != W:
            if not self._printed_resize_warning:
                logging.warning(
                    "Resizing tomogram from %s to %s", (h, w), (H, W)
                )
                self._printed_resize_warning = True
            data = np.pad(data, ((0, 0), (0, H - h), (0, W - w)), mode="edge")
            h, w = H, W
        assert (
            h % 16 == 0 and w % 16 == 0
        ), f"Invalid height: {h} or width: {w}"

        np_data = np.expand_dims(data, axis=1)  # D, C, H, W (i.e., B, C, H, W)
        np_data = np.repeat(np_data, 3, axis=1)

        torch_data = torch.from_numpy(
            np_data
        ).float()  # data expected to be float already, [0-1]
        return F.interpolate(torch_data, scale_factor=scale, mode="bicubic")



"""Module defining base data loading functionality for CryoVIT experiments."""

from abc import ABC, abstractmethod
from collections.abc import Callable
from pathlib import Path

import pandas as pd
from pytorch_lightning import LightningDataModule
from torch.utils.data import DataLoader

from cryovit.datamodules.utils import collate_fn


class BaseDataModule(LightningDataModule, ABC):
    """Base module defining common functions for creating data loaders for experiments."""

    def __init__(
        self,
        split_file: Path,
        dataset_fn: Callable,
        dataloader_fn: Callable,
        **kwargs,
    ) -> None:
        """Initializes the BaseDataModule with dataset parameters, a dataloader function, and a path to the split file.

        Args:
            split_file (Path): The path to the .csv file containing data splits.
            dataset_fn (Callable): Function to create a Dataset from a dataframe of records.
            dataloader_fn (Callable): Function to create a DataLoader from a dataset.
        """

        super().__init__()
        self.dataset_fn = dataset_fn
        self.dataloader_fn = dataloader_fn
        self.split_file = (
            split_file if isinstance(split_file, Path) else Path(split_file)
        )
        self.record_df = pd.read_csv(self.split_file)

    @abstractmethod
    def train_df(self) -> pd.DataFrame:
        """Abstract method to generate train splits."""

        raise NotImplementedError

    @abstractmethod
    def val_df(self) -> pd.DataFrame:
        """Abstract method to generate validation splits."""

        raise NotImplementedError

    @abstractmethod
    def test_df(self) -> pd.DataFrame:
        """Abstract method to generate test splits."""

        raise NotImplementedError

    @abstractmethod
    def predict_df(self) -> pd.DataFrame:
        """Abstract method to generate predict splits."""

        raise NotImplementedError

    def train_dataloader(self) -> DataLoader:
        """Creates DataLoader for training data.

        Returns:
            DataLoader: A DataLoader instance for training data.
        """

        records = self.train_df()
        if records.empty:
            raise ValueError(
                "No training data found in the provided split file."
            )
        dataset = self.dataset_fn(records, train=True)
        return self.dataloader_fn(dataset, shuffle=True, collate_fn=collate_fn)

    def val_dataloader(self) -> DataLoader:
        """Creates DataLoader for validation data.

        Returns:
            DataLoader: A DataLoader instance for validation data.
        """

        records = self.val_df()
        if records.empty:
            raise ValueError(
                "No validation data found in the provided split file."
            )
        dataset = self.dataset_fn(records, train=False)
        return self.dataloader_fn(
            dataset, shuffle=False, collate_fn=collate_fn
        )

    def test_dataloader(self) -> DataLoader:
        """Creates DataLoader for testing data.

        Returns:
            DataLoader: A DataLoader instance for testing data.
        """

        records = self.test_df()
        if records.empty:
            raise ValueError(
                "No testing data found in the provided split file."
            )
        dataset = self.dataset_fn(records, train=False)
        return self.dataloader_fn(
            dataset, shuffle=False, collate_fn=collate_fn
        )

    def predict_dataloader(self) -> DataLoader:
        """Creates DataLoader for prediction data.

        Returns:
            DataLoader: A DataLoader instance for prediction data.
        """

        records = self.predict_df()
        if records.empty:
            raise ValueError(
                "No prediction data found in the provided split file."
            )
        dataset = self.dataset_fn(records, train=False)
        return self.dataloader_fn(
            dataset, shuffle=False, collate_fn=collate_fn
        )



from pathlib import Path
from typing import Annotated

from typer import Argument, Option

from .cli import cli


@cli.command(name="infer", no_args_is_help=True)
def infer(
    tomograms: Annotated[
        str,
        Argument(
            help="Path to the folder or .txt file containing the tomograms to process.",
        ),
    ],
    model: Annotated[
        str,
        Option(
            help="Path to the .model file containing the pre-trained model.",
        ),
    ],
    result_folder: Annotated[
        str | None,
        Option(
            help="Path to the folder where the inference results will be saved.",
            show_default="the current working directory",
        ),
    ] = None,
    threshold: Annotated[
        float,
        Option(
            min=0.0,
            max=1.0,
            help="Threshold for binary segmentation.",
        ),
    ] = 0.5,
):
    """Segment tomograms using a pre-trained model.

    Example
    -------
    cryovit infer <path-to-tomograms> --model <path-to-model> --result-folder <path-to-result-folder>
    """
    from cryovit._logging_config import setup_logging
    from cryovit.run.infer_model import run_inference
    from cryovit.utils import load_files_from_path

    setup_logging("INFO")

    ## Convert Arguments
    tomograms_path = Path(tomograms)
    model_path = Path(model)
    result_path = (
        Path(result_folder) if result_folder else Path.cwd() / "predictions"
    )

    ## Sanity Checking
    assert tomograms_path.exists(), "Tomograms path does not exist."
    assert (
        model_path.exists() and model_path.suffix == ".model"
    ), "Model path does not exist or is not a .model file."
    result_path.mkdir(parents=True, exist_ok=True)

    tomogram_files = load_files_from_path(tomograms_path)

    run_inference(tomogram_files, model_path, result_path, threshold=threshold)
